# Author: lizhiyuan  <lizhiyuan@baqis.ac.cn>  2021/01/29

import functools
from qos_tools.fit import function, fitfunc as Function
from qos_tools.fit.fitclass import FitClass


def FitIndex(index='Linear'):
    if isinstance(index, str) and hasattr(Function, index):
        return getattr(Function, index)
    elif isinstance(index, str) and hasattr(function, index):
        fitfunc = getattr(function, index)
        return functools.partial(FitClass, fitfunc=fitfunc)
    elif callable(index):
        return functools.partial(FitClass, fitfunc=index)
    else:
        raise Exception(f'{index} not in FitIndex !')



# Author: lizhiyuan  <lizhiyuan@baqis.ac.cn>  2021/07/31

import itertools
import numpy as np

def CartesianSweepIterator(sweep_config={}, sweep_setting={}):
    '''Cartesian迭代函数，返回一个生成器

    Parameters:
        sweep_config: sweep_setting里key和其对应的地址以及额外的其他参数等，
            比如，{
                'name': 'S21',
                'Time': {
                    'addr': 'cfg.Q1.Gate.width',
                    'unit': 'a.u.'
                },
                'Amplitude': {
                    'addr': 'cfg.Q1.Gate.amp',
                    'unit': 'a.u.'
                },
            }
        sweep_setting: 
            比如，{
                'Time': np.linspace(0,3,4),
                'Amplitude': np.linspace(0,3,4),
            }
    '''
    for arg in itertools.product(*sweep_setting.values()):
        para_dict = {sweep_config[k]['addr']: v for k,
                   v in zip(sweep_setting.keys(), arg)}
        yield para_dict


# Author: Liu Pei  <liupei200546@163.com>  2021/08/14

from typing import Literal, Optional

import numpy as np
from matplotlib.pyplot import Axes
from pydantic import BaseModel
from scipy.fft import fft, fftfreq
from scipy.interpolate import interp1d
from scipy.signal import argrelextrema, detrend, find_peaks, savgol_filter
from scipy.stats import multivariate_normal
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import ExpSineSquared, WhiteKernel

__all__ = [
    'get_gauss_kernel', 'get_Gauss_kernel_2D', 'get_rotate_phi',
    'get_convolve_with_interpolate', 'get_convolve_arg', 'get_normalization',
    'get_background_elimination', 'get_positive_fft_abs', 'get_peaks', 'get_interpolation_peaks',
    'get_center_minimize', 'get_center_kmeans',
    'get_general_fit', 'get_expsinesquared_arg',
]


class LiteralCheck(BaseModel):
    xy: Optional[Literal['X', 'Y', 'XY']] = None
    ext: Optional[Literal['max', 'min', 'both']] = None
    mode: Optional[Literal['linear', 'constant', 'savgol']] = None


def get_general_fit(x: np.ndarray, y: np.ndarray,
                    length_scale: Optional[float] = None, periodicity: Optional[float] = None,
                    length_scale_bounds: Optional[tuple[float]] = None, periodicity_bounds: Optional[tuple[float]] = None, periodicity_count: int = 1,
                    noise_level: float = 1, noise_level_bounds: tuple[float] = (1e-5, 1e5),
                    interp_points_times: float = 5, ax: Optional[Axes] = None, **kw
                    ) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """ Author Ziyu Tao

    A general smooth method with white noise, use sklearn method.

    Args:
        x (np.ndarray): 1D array, independent variables.
        y (np.ndarray): 1D array, dependent variables.
        length_scale (Optional[float], optional): ExpSineSquared parameters. Defaults to None.
        periodicity (Optional[float], optional): ExpSineSquared parameters. Defaults to None.
        length_scale_bounds (Optional[tuple[float]], optional): ExpSineSquared parameters. Defaults to None.
        periodicity_bounds (Optional[tuple[float]], optional): ExpSineSquared parameters. Defaults to None.
        periodicity_count (int, optional): how much periods in your data. Defaults to 1.
        noise_level (float, optional): WhiteKernel parameters. Defaults to 1.
        noise_level_bounds (tuple[float], optional): WhiteKernel parameters. Defaults to (1e-5, 1e5).
        interp_points_times (float, optional): The ratio of interpolation points to original data points. Defaults to 5.
        ax (Optional[Axes], optional): the plot ax. Defaults to None.

    Returns:
        tuple[np.ndarray, np.ndarray, np.ndarray]: x data and y data after smooth with std 
    """

    length_scale = 1 if length_scale is None else length_scale
    length_scale_bounds = (
        1e-5, 1e5) if length_scale_bounds is None else length_scale_bounds
    periodicity = (np.max(x)-np.min(x)) / \
        periodicity_count if periodicity is None else periodicity
    periodicity_bounds = (
        1e-2*periodicity, 1e2*periodicity) if periodicity_bounds is None else periodicity_bounds

    gp_kernel = ExpSineSquared(length_scale=length_scale, periodicity=periodicity, length_scale_bounds=length_scale_bounds,
                               periodicity_bounds=periodicity_bounds)+WhiteKernel(noise_level=noise_level, noise_level_bounds=noise_level_bounds)

    gpr = GaussianProcessRegressor(kernel=gp_kernel)
    gpr.fit(x.reshape(-1, 1), y.reshape(-1, 1))
    xNew = np.linspace(np.min(np.real(x)), np.max(np.real(x)),
                       round(interp_points_times*x.shape[0])+1)
    yNew, yStd = gpr.predict(xNew.reshape(-1, 1), return_std=True)
    yNew = yNew.reshape([yNew.shape[0], ])
    yStd = yStd.reshape([yStd.shape[0], ])

    if ax is not None:
        ax.fill_between(xNew, yNew-yStd, yNew+yStd, alpha=0.2)
        ax.plot(x, y, '.', lw=1, ms=2)
        ax.plot(xNew, yNew, '-', lw=1, ms=2)

    return xNew, yNew, yStd


def get_expsinesquared_arg(x: np.ndarray, y: np.ndarray,
                           ext: Literal['max', 'min', 'both'] = 'min',
                           ax: Optional[Axes] = None, alpha: float = 0.1, **kw) -> tuple[bool, float]:
    """[summary]

    Args:
        x (np.ndarray): 1D array, independent variables.
        y (np.ndarray): 1D array, dependent variables.
        ext (Literal['max', 'min', 'both'], optional): extreme value type, maximum, minimum, or both. Defaults to 'min'.
        ax (Optional[Axes], optional): the plot ax. Defaults to None.
        alpha (float, optional): std upper bound. Defaults to 0.1.

    Returns:
        tuple[bool, float]: the flag of extreme greater than several times of the standard deviation and the extreme position.
    """

    ext = LiteralCheck(ext=ext).ext
    xNew, yNew, yStd = get_general_fit(x=x, y=y, ax=ax, **kw)

    x_min_pos, = (np.array([], dtype=np.int64),
                  ) if ext == 'max' else argrelextrema(yNew, np.less)
    x_max_pos, = (np.array([], dtype=np.int64),
                  ) if ext == 'min' else argrelextrema(-yNew, np.less)
    x_ans = np.concatenate((x_min_pos, x_max_pos))

    if len(x_ans) == 0:
        x_ans = np.array([yNew.shape[0]//2], dtype=np.int64)

    if ax is not None:
        ax.axvline(x=xNew[x_ans[0]], lw=1, ls='-')

    return yStd[x_ans[0]] < alpha, xNew[x_ans[0]]


def get_error_output(data: float, err: float) -> tuple[float, int, int]:
    """
    Error analysis

    Args:
        data (float): data
        err (float): error

    Returns:
        tuple[float, int, int]: data, err, significant digits
    """

    ret = 0
    if np.isnan(err) or err>1:
        return data, 0, 0
    while round(err) < 1:
        ret += 1
        err = err*10
    return round(data*(10**ret))/10**ret, round(err), ret


def get_center_minimize(S: np.ndarray, N: int) -> np.ndarray:
    """
    Using a simple nearest method to determin centers of clusters on IQ plane.

    Args:
        S (np.ndarray): data, in the shape [n, shots], n is number of groups initially classified and shots is the all data, here all the data n*shots are used in classification algorithm.
        N (int): number of clusters

    Returns:
        np.ndarray: array of centers in complex
    """

    def get_center(x, S, N):
        minS = np.ones([S.shape[0], N])
        for i in range(N):
            minS[:, i] = np.real(np.abs(S-x[i]-1j*x[i+N]))
        minS = np.min(minS, axis=1)
        return np.sum(minS)

    from scipy.optimize import minimize

    x0 = []

    for i in range(N):
        x0.append(np.real(np.mean(S[i])))
    for i in range(N):
        x0.append(np.imag(np.mean(S[i])))

    result = minimize(get_center, x0=x0, args=(S.flatten(), N))

    cList = []
    for i in range(N):
        cList.append(result.x[i]+1j*result.x[i+N])
    cList = np.asarray(cList)
    return cList


def get_center_kmeans(S: np.ndarray, N: int):
    """
    Using kmeans method to determin centers of clusters on IQ plane.

    Args:
        S (np.ndarray): data, in the shape [n, shots], n is number of groups initially classified and shots is the all data, here all the data n*shots are used in classification algorithm.
        N (int): number of clusters

    Returns:
        np.ndarray: array of centers in complex
    """

    def exchange(x, center):
        from itertools import permutations
        assert x.shape[0] == center.shape[0], 'n not equal'
        permutation = list(permutations(range(x.shape[0]), x.shape[0]))
        cost = np.array([np.sum(np.abs(x-center[list(item)])**2)
                        for item in permutation])
        return center[list(permutation[np.argmin(cost)])]

    from sklearn.cluster import KMeans

    SS = S.flatten()
    classification = KMeans(n_clusters=N).fit_predict(
        np.array([np.real(SS), np.imag(SS)]).T)

    clist = []
    for i in range(N):
        clist.append(np.mean(SS[np.where(classification == i)]))
    return exchange(np.array([np.mean(S[i]) for i in range(N)]), np.asarray(clist))


def get_gauss_kernel(halfsize: int = 1, a: float = 2) -> np.ndarray:
    """
    Return the one-dimensional Gaussian convolution kernel function.

    Args:
        halfsize (int, optional): half side of th kernel function and total length is 2*halfsize+1. Defaults to 1.
        a (float, optional): the ratio of the range of values of the Gaussian distribution function to the variance. Defaults to 2.

    Returns:
        np.ndarray: 1D Gaussian convolution kernel function
    """

    rv = multivariate_normal(0, 1)
    x0 = np.linspace(-a, a, 2*halfsize+1)
    _m = rv.pdf(x0)
    m = _m/np.sum(_m)
    return m


def get_Gauss_kernel_2D(halfsize: int = 2, a: float = 2, factor: float = 1,
                        xy: Literal['X', 'Y'] = 'X') -> np.ndarray:
    """
    Return the two-dimensional Gaussian convolution kernel matrix.

    Args:
        halfsize (int, optional): half side of th kernel function and total length is 2*halfsize+1. Defaults to 1.
        a (float, optional): the ratio of the range of values of the Gaussian distribution function to the variance. Defaults to 2.
        factor (float, optional): covariance factor of a two-dimensional Gaussian function. Defaults to 1.
        xy (Literal, optional): control covariance coefficient pattern. Defaults to 'X'.
    Returns:
        np.ndarray: 2D Gaussian convolution kernel matrix。
    """

    xy = LiteralCheck(xy=xy).xy

    mean = [0, 0]
    if xy == 'X':
        cov = [[1, 0], [0, factor]]
    elif xy == 'Y':
        cov = [[factor, 0], [0, 1]]
    else:
        cov = [[1, 0], [0, 1]]
    rv = multivariate_normal(mean, cov)
    x0 = np.linspace(-a, a, 2*halfsize+1)
    x, y = np.meshgrid(x0, x0)
    pos = np.dstack((x, y))
    _m = rv.pdf(pos)
    m = _m/np.sum(_m)
    return m


def get_rotate_phi(x: np.ndarray,
                   ax: Optional[Axes] = None) -> np.ndarray:
    """
    Rotate points on a straight line in the plane onto the X-axis. The point near the origin is going to be on the left.

    Args:
        x (Optional[np.ndarray]): complex data
        ax (Optional[Axes], optional): the plot ax. Defaults to None.

    Returns:
        np.ndarray: the complex data after processed.
    """

    xx = np.ndarray.flatten(x)

    func = np.poly1d(np.polyfit(x=np.real(xx), y=np.imag(xx), deg=1))

    if ax is not None:
        ax.plot(np.real(x), np.imag(x), '.', lw=1, ms=2)
        xNew = np.linspace(np.min(np.real(x)), np.max(np.real(x)), 1001)
        ax.plot(np.real(xNew), func(np.real(xNew)), '-', lw=1, ms=2)

    xmin, xmax = np.argmin(np.real(xx)), np.argmax(np.real(xx))
    slope = (func(np.max(np.real(xx)))-func(np.min(np.real(xx)))) / \
        (np.max(np.real(xx))-np.min(np.real(xx)))

    if np.abs(xx[xmin]) > np.abs(xx[xmax]):
        slope *= -1

    if ax is not None:
        xxx = np.real(x*(1-1j*slope)/np.sqrt(1+slope**2))
        ax.plot(np.real(xxx), np.imag(xxx), '.', lw=1, ms=2)
        ax.plot(np.real(xxx[0]), np.imag(xxx[0]), '.', lw=1, ms=5)

    return np.real(x*(1-1j*slope)/np.sqrt(1+slope**2))


def get_convolve_with_interpolate(x: np.ndarray, y: np.ndarray,
                                  ker: np.ndarray = np.array(
                                      [0.25, 0.5, 0.25]),
                                  interp_points_times: float = 3, kind: str = 'cubic',
                                  ax: Optional[Axes] = None) -> tuple[np.ndarray, np.ndarray]:
    """
    Return y's convolve with interpolation, that is, interpolate the result of the convolution.

    Args:
        x (np.ndarray): ndarray of independent variables.
        y (np.ndarray): ndarray of dependent variables.
        ker (np.ndarray, optional): convolution kernel function. Defaults to np.array([0.25, 0.5, 0.25]).
        interp_points_times (float, optional): The ratio of interpolation points to original data points. Defaults to 3.
        kind (str, optional): interpolation function type, which is used in `numpy.interp1d`. Defaults to 'cubic'.
        ax (Optional[Axes], optional): the plot ax. Defaults to None.

    Returns:
        tuple[np.ndarray, np.ndarray]: x and y after processed.
    """

    convolve_y = np.convolve(y, ker, mode='valid')
    x_offset = (ker.shape[0])//2
    xx = x[x_offset:x_offset+convolve_y.shape[0]]
    func = interp1d(xx, convolve_y, kind=kind)
    xNew = np.linspace(np.min(xx), np.max(xx), round(
        convolve_y.shape[0]*interp_points_times)+1)
    yNew = func(xNew)

    if ax is not None:
        ax.plot(x, y, '.', lw=1, ms=2)
        ax.plot(xNew, yNew, '-', lw=1, ms=2)

    return xNew, yNew


def get_convolve_arg(x: np.ndarray, y: np.ndarray,
                     ker: np.ndarray = np.array([0.25, 0.5, 0.25]),
                     alpha: float = 1, ext: Literal['max', 'min', 'both'] = 'min',
                     interp_points_times: float = 3,
                     ax: Optional[Axes] = None, kind: str = 'cubic') -> tuple[bool, float]:
    """
    Find argmin of argmax of y after a simple convolve with interpolate and return the corresponding item in independent variables

    Args:
        x (np.ndarray): ndarray of independent variables.
        y (np.ndarray): ndarray of dependent variables.
        ker (np.ndarray, optional): convolution kernel function. Defaults to np.array([0.25, 0.5, 0.25]).
        alpha (float, optional): the ratio of the difference between extreme value and mean value to standard deviation. Defaults to 1.
        ext (Literal['max', 'min', 'both'], optional): extreme value type, maximum, minimum, or both. Defaults to 'min'.
        interp_points_times (float, optional): The ratio of interpolation points to original data points. Defaults to 3.
        ax (Axes, optional): the plot ax. Defaults to None.
        kind (str, optional): interpolation function type, which is used in `numpy.interp1d`. Defaults to 'cubic'.

    Returns:
        tuple[bool, float]: the flag of extreme greater than several times of the standard deviation and the extreme position.
    """

    ext = LiteralCheck(ext=ext).ext

    xNew, yNew = get_convolve_with_interpolate(
        x=x, y=y, ker=ker, interp_points_times=interp_points_times, kind=kind, ax=ax)
    std_y = np.std(yNew)

    pos_x, dif_y = (xNew[np.argmax(yNew)], np.abs(np.max(yNew)-np.mean(yNew))) if ext in [
        'max'] else (xNew[np.argmin(yNew)], np.abs(np.min(yNew)-np.mean(yNew)))

    if ax is not None:
        ax.axvline(x=pos_x, lw=1, ls='-')

    return dif_y > std_y*alpha, pos_x


def get_normalization(x: np.ndarray) -> np.ndarray:
    """
    Normalize a array to [0, 1].

    Args:
        x (np.ndarray): a array.

    Returns:
        np.ndarray: the processed array
    """
    return (x-np.min(x))/(1 if np.max(x) == np.min(x) else (np.max(x)-np.min(x)))


def get_background_elimination(x: np.ndarray,
                               mode: Literal['linear',
                                             'constant', 'savgol'] = 'linear',
                               savgol_times: Optional[float] = None, savgol_order: int = 1) -> np.ndarray:
    """
    Return data removed the background of 1D data array.

    Args:
        y (np.ndarray): the data.
        mode (str, optional): in ['linear'(for deducting linear trend), 'constant'(for deducting a constant), 'savgol'(for savgol filter)]. Defaults to 'savgol'.
        savgol_times (Optional[float], optional): the parameters for savgol filter. Defaults to 0.9.
        savgol_order (int, optional): the parameters for savgol filter. Defaults to 1.

    Returns:
        np.ndarray: data after deducting the background
    """

    mode = LiteralCheck(mode=mode).mode

    if mode in ['linear', 'constant']:
        return detrend(x, type=mode)
    else:
        assert savgol_times is not None, 'Please input `savgol_times`.'
        return x-savgol_filter(x, window_length=round(x.shape[0]*savgol_times//2)*2+1, polyorder=savgol_order)


def get_positive_fft_abs(x: np.ndarray, y: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    """
    Return the positive frequency of fft.

    Args:
        x (np.array): ndarray of independent variables.
        y (np.array): ndarray of dependent variables.

    Raises:
        ValueError: shape error.

    Returns:
        tuple[np.array, np.array]: fft frequency, fft amplitude.
    """

    if x.shape[0] != y.shape[0] or x.shape[0] == 0 or y.shape[0] == 0:
        raise ValueError('x, y must have the same length or someone is null')
    return fftfreq(x.shape[0], x[1]-x[0])[1:(x.shape[0]//2)], (get_normalization(np.abs(fft(y))))[1:(x.shape[0]//2)]


def get_peaks(x: np.ndarray,
              height: Optional[float] = None, peak_height_times: float = 0.5,
              distance_times: float = 0.1) -> tuple[np.ndarray, dict]:
    """
    Return peaks position.

    Args:
        x (np.ndarray): ndarray of independent variables.
        height (Optional[float], optional): the height in finding peak. Defaults to None.
        peak_height_times (float, optional): ratio of the lowerbound height of a peak to the maximum difference. Defaults to 0.5.
        distance_times (float, optional): ratio of the distance to the number of points between the peaks. Defaults to 0.1.

    Returns:
        tuple[np.ndarray, dict]: [description]
    """

    return find_peaks(x,
                      height=(np.max(x)-np.min(x)) *
                      peak_height_times if height is None else height,
                      distance=max(1, round(distance_times*x.shape[0])))


def get_interpolation_peaks(x: np.ndarray, y: np.ndarray,
                            ext: Literal['max', 'min', 'both'] = 'max',
                            kind: str = 'cubic', interp_points_times: float = 3,
                            savgol_length_times: Optional[float] = None, savgol_order: int = 1,
                            height: Optional[float] = None, peak_height_times: float = 0.5,
                            distance_times: float = 0.1, ax: Optional[Axes] = None) -> list[float]:
    """
    Return peaks after a simple interpolation and background removed, which a savgol filter can be used or not.

    Args:
        x (np.ndarray): ndarray of independent variables.
        y (np.ndarray): ndarray of dependent variables.
        ext (Literal['max', 'min', 'both'], optional): extreme value type, maximum, minimum, or both. Defaults to 'max'.
        kind (str, optional): interpolation function type, which is used in `numpy.interp1d`. Defaults to 'cubic'.
        interp_points_times (float, optional): The ratio of interpolation points to original data points. Defaults to 3.
        height (Optional[float], optional): the height in finding peak. Defaults to None.
        peak_height_times (float, optional): ratio of the lowerbound height of a peak to the maximum difference. Defaults to 0.5.
        distance_times (float, optional): ratio of the distance to the number of points between the peaks. Defaults to 0.1.
        ax (Optional[Axes], optional): the plot ax. Defaults to None.
        savgol_times (Optional[float], optional): the parameters for savgol filter. Defaults to None.
        savgol_order (int, optional): the parameters for savgol filter. Defaults to 1.

    Returns:
        list[float]: the corresponding value of the peaks position.
    """

    ext = LiteralCheck(ext=ext).ext

    ydata = get_background_elimination(y, mode='linear')
    if ext in ['min']:
        ydata = -ydata
    elif ext in ['both']:
        ydata = np.abs(ydata)

    func = interp1d(x, ydata, kind=kind)
    xdata = np.linspace(np.min(x), np.max(
        x), interp_points_times*ydata.shape[0]+1)
    ydata = func(xdata) if savgol_length_times is None else savgol_filter(func(xdata),
                                                                          window_length=int(savgol_length_times*len(xdata)/2)*2+1, polyorder=savgol_order)
    peak_position, _ = get_peaks(
        x=ydata, height=height, peak_height_times=peak_height_times, distance_times=distance_times)

    if ax is not None:
        ax.plot(x, y, '.-', lw=1, ms=2)
        ax.plot(xdata, ydata, '.-', lw=1, ms=2)
        if len(peak_position) > 0:
            for item in xdata[peak_position]:
                ax.axvline(x=item, lw=1, ls='-')
    return xdata[peak_position]


def get_gauss_fit(x: np.ndarray, ax: Optional[Axes] = None, sigma=3, fit=True, **kw):
    def _erf(x, mu, sigma):
        from scipy.special import erf
        return 0.5+0.5*erf(((x-mu)/sigma)/np.sqrt(2))

    from scipy.optimize import curve_fit
    xx, yy = np.sort(x), (np.arange(x.shape[0])+1)/x.shape[0]
    if not fit:
        popt, pcov = np.median(xx), np.std(xx)
        if ax is not None:
            ax.plot(xx, yy, '.-', **kw)
            ax.axvline(x=popt, c=kw.get('c', None))
    else:
        popt, pcov = curve_fit(_erf, xx, yy, p0=[np.mean(x), np.std(x)])
        if ax is not None:
            ax.plot(xx, yy, '.', **kw)
            xxx = np.linspace(popt[0]-popt[1]*sigma, popt[0]+popt[1]*sigma, 1001)
            if 'label' in kw:
                kw.pop('label')
            ax.plot(xxx, _erf(xxx, *popt), '-', **kw)
    
    return popt, pcov



# Author: lizhiyuan  <lizhiyuan@baqis.ac.cn>  2021/08/08
from qos_tools.gatemap.__info_extract import extract_align_info, extract_gateMap_priority, extract_priority_info, extract_gate_info


def stdgatemap2stdcmds(stdgatemap):
    '''
    stdgatemap: 形如，
        {
            'Q1': ['X', 'X/2', ':1:', 'CZ(Q1,Q2)', 'I', 'Y', ':3:'],
            'Q2': ['-Y/2',':1:','CZ(Q1,Q2)',':4:','CZ(Q2,Q3)','I','Y',':3:',],
            'Q3': [':4:', 'CZ(Q2,Q3)', 'I', 'Y', ':3:', 'Y'],
        }
    stdcmds: 形如，
        [('-Y/2', ('Q2',), ()),
        ('X', ('Q1',), ()),
        ('X/2', ('Q1',), ()),
        ('Barrier', ('Q1', 'Q2'), ()),
        ('CZ', ('Q1', 'Q2'), ()),
        ('I', ('Q1',), ()),
        ('Barrier', ('Q2', 'Q3'), ()),
        ('CZ', ('Q2', 'Q3'), ()),
        ('Y', ('Q1',), ()),
        ('I', ('Q2',), ()),
        ('I', ('Q3',), ()),
        ('Y', ('Q3',), ()),
        ('Y', ('Q2',), ()),
        ('Barrier', ('Q1', 'Q2', 'Q3'), ()),
        ('Y', ('Q3',), ())]
    '''
    align_info = extract_align_info(stdgatemap)
    gateMap_priority = extract_gateMap_priority(stdgatemap, align_info)
    priority_info = extract_priority_info(gateMap_priority)
    stdcmds = extract_gate_info(stdgatemap, priority_info)
    return stdcmds


def stdcmds2stdgatemap(stdcmds):
    stdgatemap = {}
    count = 0
    for operator, qubits, params in stdcmds:
        for qubit in qubits:
            if qubit not in stdgatemap.keys():
                stdgatemap[qubit] = []

        if len(qubits) == 1:
            qubit, = qubits
            stdindex = (operator, (), params)
            stdgatemap[qubit].append(stdindex)
        else:
            if operator in ['Barrier']:
                stdindex = None
            else:
                stdindex = (operator, qubits, params)
            alignTag = (f'Barrier:{count}', tuple(), tuple())
            count += 1
            for qubit in qubits:
                stdgatemap[qubit].append(alignTag)
                if stdindex is not None:
                    stdgatemap[qubit].append(stdindex)
    return stdgatemap


def gatemap_convert(gatemap, fr='stdindex', to='tupindex'):
    from ._index_convert import index_convert
    res_gatemap = {}
    for q, gatelist in gatemap.items():
        if q not in res_gatemap.keys():
            res_gatemap[q] = []

        for index in gatelist:
            _index = index_convert(index, fr, to)
            res_gatemap[q].append(_index)
    return res_gatemap


def cmds_convert(cmds, fr='stdcmd', to='qlispcmd'):
    from ._cmd_convert import cmd_convert
    res_cmds = []
    for cmd in cmds:
        _cmd = cmd_convert(cmd, fr, to)
        res_cmds.append(_cmd)
    return res_cmds


def circuit_convert(circuit, fr='gatemap:tupindex', to='cmds:qlispcmd'):
    ''''''
    if fr != to:
        cir_type_fr, cmd_type_fr = fr.split(':')
        cir_type_to, cmd_type_to = to.split(':')

        cir_type_support = ['gatemap', 'cmds']
        assert cir_type_fr in cir_type_support, f"circuit tpye '{cir_type_fr}' not support!"
        assert cir_type_to in cir_type_support, f"circuit tpye '{cir_type_to}' not support!"

        cmd_type_support = ['strindex', 'tupindex', 'stdindex',
                            'stdcmd', 'qlispcmd']
        assert cmd_type_fr in cmd_type_support, f"cmd tpye '{cmd_type_fr}' not support!"
        assert cmd_type_to in cmd_type_support, f"cmd tpye '{cmd_type_to}' not support!"

        if cir_type_fr in ['gatemap']:
            circuit = gatemap_convert(circuit, cmd_type_fr, 'stdindex')
        elif cir_type_fr in ['cmds']:
            circuit = cmds_convert(circuit, cmd_type_fr, 'stdcmd')
        else:
            raise

        if cir_type_fr == cir_type_to:
            pass
        else:
            if cir_type_fr == 'gatemap' and cir_type_to == 'cmds':
                circuit = stdgatemap2stdcmds(circuit)
            elif cir_type_fr == 'cmds' and cir_type_to == 'gatemap':
                circuit = stdcmds2stdgatemap(circuit)
            else:
                raise

        if cir_type_to in ['gatemap']:
            circuit = gatemap_convert(circuit, 'stdindex', cmd_type_to)
        elif cir_type_to in ['cmds']:
            circuit = cmds_convert(circuit, 'stdcmd', cmd_type_to)
        else:
            raise
    return circuit



# Author: lizhiyuan  <lizhiyuan@baqis.ac.cn>  2021/01/01

__all__ = [
    'flatten_dict', 'restore_dict',
]


def flatten_dict(d, symbol='.', level=100):
    '''将多层字典压平为一层
    Parameters:
        d: 待压平的字典
        symbol: 压平所用的连接符号
        level: 压平字典的级数，0为不做操作，默认100级
    Return:
        压平后的字典
    '''
    fd = {}
    for k, v in d.items():
        if isinstance(v, dict) and bool(v) and level > 0:  # v 非空字典
            fd1 = flatten_dict(v, symbol, level-1)
            fd2 = dict(zip((k+symbol+_k for _k in fd1.keys()), fd1.values()))
            fd.update(fd2)
        else:
            fd.update({k: v})
    return fd


def restore_dict(d, symbol='.'):
    '''上面 flatten_dict 函数的逆过程，将压平的字典还原'''
    rd = {}
    for k, v in d.items():
        ks = k.split(symbol)
        _d = rd
        for _k in ks[:-1]:
            _d.setdefault(_k, {})
            _d = _d.get(_k)
        _d.update({ks[-1]: v})
    return rd


def check_dict_template(d, template):
    '''检查字典d是否包含模板字典template中所有的关键字'''
    return flatten_dict(d).keys() >= flatten_dict(template).keys()



# Author: lizhiyuan  <lizhiyuan@baqis.ac.cn>  2021/08/17
from graphlib import TopologicalSorter

__all__ = [
    'graph_reverse',
    'get_subgraph',
    'get_all_nodes',
    'get_topo_order',
    'get_node_priority',
    'get_priority_node',
    'get_exec_order',
    'extract_graph',
]


def graph_reverse(graph_forward):
    '''
    Parameter:
        graph_forward: dict，参考 graphlib.TopologicalSorter 参数graph的定义，如下
                    graph argument must be a dictionary representing a directed acyclic graph 
                    where the keys are nodes and the values are iterables of all predecessors of that node in the graph 
                    (the nodes that have edges that point to the value in the key).
                    graph = {"D": {"B", "C"}, "C": {"A"}, "B": {"A"}}
                    其中key是节点，value是此节点的前置节点。
    Return：
        graph_backward: 反转逆向的图
    '''
    graph_backward = {}
    for k, v in graph_forward.items():
        for _v in v:
            if _v not in graph_backward.keys():
                graph_backward[_v] = set()
            graph_backward[_v].add(k)
    return graph_backward


def get_subgraph(keys=None, graph=None, graph_backward=None, _new_graph=None):
    '''
    Parameter:
        keys: list/tuple/set/iterable, 子图的起始节点或包含在子图中的节点
        graph: dict，参考 graphlib.TopologicalSorter 参数graph的定义，
            graph = {"D": {"B", "C"}, "C": {"A"}, "B": {"A"}}
            其中key是节点，value是此节点的前置节点。
        graph_backward: graph的反向图，graph和graph_backward二选一，优先使用graph_backward
        _new_graph: 用于递归的中间参量
    Return：
        _new_graph: 新生成的子图
    '''
    keys = set() if keys is None else keys
    graph = {} if graph is None else graph
    _new_graph = {} if _new_graph is None else _new_graph
    graph_backward = graph_reverse(
        graph) if graph_backward is None else graph_backward

    next_add_keys = set()
    for k in keys:
        next_nodes = graph_backward.get(k, set())
        next_add_keys.update(next_nodes)
        for next_node in next_nodes:
            if next_node not in _new_graph.keys():
                _new_graph[next_node] = set()
            _new_graph[next_node].add(k)

    if next_add_keys:
        _new_graph = get_subgraph(
            next_add_keys, graph_backward=graph_backward, _new_graph=_new_graph)
    return _new_graph


def get_all_nodes(graph):
    '''
    Parameter:
        graph: dict，参考 graphlib.TopologicalSorter 参数graph的定义，
            graph = {"D": {"B", "C"}, "C": {"A"}, "B": {"A"}}
            其中key是节点，value是此节点的前置节点。
    Return：
        all_nodes: set，图所有节点的集合，比如
            {'A', 'B', 'C', 'D'}
    '''
    all_nodes = set()
    for k, v in graph.items():
        all_nodes.add(k)
        all_nodes.update(v)
    return all_nodes


def get_topo_order(graph):
    '''
    Parameter:
        graph: dict，参考 graphlib.TopologicalSorter 参数graph的定义，
            graph = {"D": {"B", "C"}, "C": {"A"}, "B": {"A"}}
            其中key是节点，value是此节点的前置节点。
    Return：
        topo_order: list，按照拓扑序排列的节点列表，比如
            ['A', 'B', 'C', 'D']
    '''
    ts = TopologicalSorter(graph)
    return list(ts.static_order())


def get_node_priority(graph):
    '''
    Parameter:
        graph: dict，参考 graphlib.TopologicalSorter 参数graph的定义，
            graph = {"D": {"B", "C"}, "C": {"A"}, "B": {"A"}}
            其中key是节点，value是此节点的前置节点。
    Return：
        node_priority: dict，节点和优先度组成的键值对字典，比如
            {'A': 1, 'C': 2, 'B': 2, 'D': 3}
    '''
    ts = TopologicalSorter(graph)
    node_priority = {}
    for k in ts.static_order():
        prenode_priority = [node_priority[_k] for _k in graph.get(k, set())]
        max_prenode = max(prenode_priority, default=0)
        node_priority[k] = max_prenode+1
    return node_priority


def get_priority_node(graph):
    '''
    Parameter:
        graph: dict，参考 graphlib.TopologicalSorter 参数graph的定义，
            graph = {"D": {"B", "C"}, "C": {"A"}, "B": {"A"}}
            其中key是节点，value是此节点的前置节点。
    Return：
        priority_node: dict，优先度和节点组成的键值对字典，比如
            {1: {'A'}, 2: {'B', 'C'}, 3: {'D'}}
    '''
    node_priority = get_node_priority(graph)
    priority_node = {}
    for node, priority in node_priority.items():
        if priority not in priority_node.keys():
            priority_node[priority] = set()
        priority_node[priority].add(node)
    return priority_node


def get_exec_order(graph):
    '''
    Parameter:
        graph: dict，参考 graphlib.TopologicalSorter 参数graph的定义，
            graph = {"D": {"B", "C"}, "C": {"A"}, "B": {"A"}}
            其中key是节点，value是此节点的前置节点。
    Return：
        exec_order: list，按照优先度排序的节点序列，比如
            [{'A'}, {'B', 'C'}, {'D'}]
    '''
    priority_node = get_priority_node(graph)
    exec_order = [priority_node[p+1] for p in range(len(priority_node.keys()))]
    return exec_order


def extract_graph(constrains):
    '''
    Parameter:
        constrains: tuple(deps,func,targets)，其中 deps/targets 为 str|tuple(str)
    Return：
        graph_forward, graph_backward: 前向图和反向图
    '''
    graph_forward = {}
    graph_backward = {}

    for deps, func, targets in constrains:
        if isinstance(deps, str):
            deps = (deps,)
        if isinstance(targets, str):
            targets = (targets,)

        for target in targets:
            if target not in graph_forward.keys():
                graph_forward[target] = set()
            graph_forward[target].update(deps)

        for dep in deps:
            if dep not in graph_backward.keys():
                graph_backward[dep] = set()
            graph_backward[dep].update(targets)

    return graph_forward, graph_backward



# Author: Liu Pei  <liupei200546@163.com>  2021/08/14
# Rebuild: Liu Pei  <liupei200546@163.com>  2022/03/16

from .create import create_qubit, create_coupler, create_probe, create_dev, create_gate_measure, create_gate_CR, create_gate_rfUnitary, create_jpa
from .map import Mapping, config2mapping, mapping2cfg


# Rebuild: Liu Pei  <liupei200546@163.com>  2022/03/29

import numpy as np
from itertools import chain
import kernel

from typing import Optional

from .tools import generate_log_intlist

__all__ = [
    'Count_n', 'AllXY_without_constrains',
    'RB_single_qubit',
    'RB_two_qubits', 'RB_two_qubits_IBM',
    'QPT_two_qubits'
]


def Count_n(qubits: list[str],
            max_n: int = 100, step_n: float = 10,
            signal: str = 'iq_avg', act_qubits: Optional[list[str]] = None, **kw) -> dict:
    """
    ['n'] Count of X gate.

    Args:
        qubits (list[str]): qubit names.
        max_n (int, optional): maximum of n. Defaults to 100.
        step_n (float, optional): step of n. Defaults to 10.
        signal (str, optional): signal. Defaults to 'iq_avg'.
        act_qubits (Optional[list[str]], optional): act qubits. Defaults to None.
    """

    act_qubits = qubits if act_qubits is None else act_qubits

    return {
        'init': {
            'name': 'N Count',
            'qubits': qubits,
            'signal': signal,
        },
        'sweep_setting': {
            'n': np.arange(step_n, max_n+1, step_n),
        },
        'sweep_addition': {
            'circuit': lambda n: [
                *[(('rfUnitary', np.pi, 0), q) for q in act_qubits]*n,
                ('Barrier', tuple(qubits)),
                *[(('Measure', j), q) for j, q in enumerate(qubits)],
            ],
        },
    }


def AllXY_without_constrains(qubits: list[str],
                             n: int = 0, repeat: int = 1,
                             signal: str = 'population', act_qubits: Optional[list[str]] = None, **kw) -> dict:
    """
    [21] AllXY 21 circuit.

    Args:
        qubits (list[str]): qubit names.
        n (int, optional): repeat of gate. Defaults to 0.
        repeat (int, optional): repeat times. Defaults to 1.
        signal (str, optional): signal. Defaults to 'population'.
        act_qubits (Optional[list[str]], optional): act qubits. Defaults to None.
    """

    def _changeI(q):
        params = kernel.get(f'gate.rfUnitary.{q}.params')
        return params.get('duration', [[0, 0.5], [0, 0]])[-1][-1]+params.get('buffer', 0)

    def _span_n(x, n, changeI: float = None):
        if n == 0:
            return x
        ret = []
        for item in x:
            n_pre = 4*n+1
            if item[0] == 'I' and changeI is not None:
                item = list(item)
                item[0] = ('Delay', changeI)
                item = tuple(item)
            if item[0] in ['X']:
                item = list(item)
                item[0] = ('rfUnitary', np.pi, 0)
                item = tuple(item)
                n_pre = 2*n+1
            if item[0] in ['Y']:
                item = list(item)
                item[0] = ('rfUnitary', np.pi, np.pi/2)
                item = tuple(item)
                n_pre = 2*n+1
            if item[0] in ['X/2']:
                item = list(item)
                item[0] = ('rfUnitary', np.pi/2, 0)
                item = tuple(item)
            if item[0] in ['Y/2']:
                item = list(item)
                item[0] = ('rfUnitary', np.pi/2, np.pi/2)
                item = tuple(item)
            ret.extend([item]*n_pre)
        return ret

    from waveforms.quantum import ALLXY

    act_qubits = qubits if act_qubits is None else act_qubits
    rfUnitary_time = {q: _changeI(q) for q in act_qubits}

    return {
        'init': {
            'name': 'All XY',
            'qubits': qubits,
            'signal': signal,
        },
        'sweep_setting': {
            'number': np.arange(21),
            'repeat': np.arange(repeat),
        },
        'sweep_addition': {
            'circuit': lambda number: [
                *chain.from_iterable(_span_n(ALLXY(qubit=q, i=number), n,
                                             rfUnitary_time[q]) for q in act_qubits),
                ('Barrier', tuple(qubits)),
                *[(('Measure', j), q) for j, q in enumerate(qubits)],
            ],
        },
    }


def RB_single_qubit(qubits: list[str],
                    start: int = 1, max_cycle: int = 500, sweep_points: int = 21,
                    random_times: int = 20,
                    base: Optional[list[list[tuple]]] = None, interleaves: list = [[]],
                    ini_gate: list = ['X'],
                    signal: str = 'population', act_qubits: Optional[list[str]] = None, **kw) -> dict:
    """
    ['seed'] RB on single qubit.

    Args:
        qubits (list[str]): qubit names.
        start (int, optional): start number. Defaults to 1.
        max_cycle (int, optional): maximum of cycle. Defaults to 500.
        sweep_points (int, optional): sweep points. Defaults to 21.
        random_times (int, optional): random times. Defaults to 20.
        base (Optional[list[list[tuple]]], optional): RB base. Defaults to None.
        interleaves (list): interleaves gate. Defaults to [].
        ini_gate (list): initial gate. Defaults to ['X'].
        signal (str, optional): signal. Defaults to 'population'.
        act_qubits (Optional[list[str]], optional): act qubits. Defaults to None.
    """

    from waveforms.quantum.rb import generateRBCircuit

    act_qubits = qubits if act_qubits is None else act_qubits

    ret = {
        'init': {
            'name': 'RB Single Qubit',
            'qubits': qubits,
            'signal': signal,
        },
        'sweep_setting': {
            'seed': [np.random.randint(0xfffffff) for _ in range(random_times)],
            'cycle': generate_log_intlist(min_int=start, max_int=max_cycle, sweep_points=sweep_points),
        },
        'sweep_addition': {
            'circuit': lambda cycle, seed, interleave_id=0, interleaves=interleaves: [
                *[(gate, q) for gate in ini_gate for q in act_qubits],
                *chain.from_iterable(generateRBCircuit(qubits=(q,), cycle=cycle,
                                                       seed=seed, interleaves=interleaves[interleave_id], base=base) for q in act_qubits),
                ('Barrier', tuple(qubits)),
                *[(('Measure', j), q) for j, q in enumerate(qubits)],
            ],
        },
    }
    
    if len(interleaves):
        ret['sweep_setting']['interleave_id'] = np.arange(len(interleaves))

    return ret


def RB_two_qubits(qubits: list[tuple[str]],
                  start: int = 1, max_cycle: int = 50, sweep_points: int = 21,
                  random_times: int = 20,
                  base: Optional[list[list[tuple]]] = None, interleaves: list = [[]],
                  ini_gate: list = [('X', 'X')],
                  signal: str = 'diag', act_qubits: Optional[list[tuple[str]]] = None, **kw) -> dict:
    """
    ['seed'] RB on two qubit.

    Args:
        qubits (list[tuple[str]]): qubit names.
        start (int, optional): start number. Defaults to 1.
        max_cycle (int, optional): maximum of cycle. Defaults to 500.
        sweep_points (int, optional): sweep points. Defaults to 21.
        random_times (int, optional): random times. Defaults to 20.
        base (Optional[list[list[tuple]]], optional): RB base. Defaults to None.
        interleaves (list): interleaves gate. Defaults to [].
        ini_gate (list, optional): initial gate. Defaults to [('X', 'X')].
        signal (str, optional): signal. Defaults to 'diag'.
        act_qubits (Optional[list[str]], optional): act qubits. Defaults to None.
    """

    act_qubits = qubits if act_qubits is None else act_qubits
    from waveforms.quantum.rb import generateRBCircuit

    ret = {
        'init': {
            'name': 'RB Two Qubits',
            'qubits': qubits,
            'signal': signal,
        },
        'sweep_setting': {
            'seed': np.random.randint(0xfffffff, size=[random_times]),
            'cycle': generate_log_intlist(min_int=start, max_int=max_cycle, sweep_points=sweep_points),
        },
        'sweep_addition': {
            'circuit': lambda cycle, seed, interleave_id=0, interleaves=interleaves: [
                *[(gate[i], q[i]) for i in range(2)
                  for gate in ini_gate for q in act_qubits],
                *chain.from_iterable(generateRBCircuit(qubits=q, cycle=cycle, interleaves=interleaves[interleave_id],
                                                       seed=seed, base=base) for q in act_qubits),
                ('Barrier', tuple(chain.from_iterable(qubits))),
                *[(('Measure', i), q) for i, q in enumerate(chain(*qubits))],
            ],
        },
    }

    if len(interleaves):
        ret['sweep_setting']['interleave_id'] = np.arange(len(interleaves))
    
    return ret


def _cr_2RB(qubits, cycle, seed, base, interleaves):

    from waveforms.quantum.clifford.clifford import inv, mul
    import random
    # from waveforms.quantum.rb import replace_qubit, index_to_circuit
    from waveforms.quantum.rb import index_to_circuit
    from waveforms.qlisp import mapping_qubits

    assert len(qubits) == 2, 'Invalid qubits length'

    MAX = 24*24

    ret = []
    index = 0
    rng = random.Random(seed)
    int_id, int_gate = interleaves

    for _ in range(cycle):
        i = rng.randrange(MAX)
        ret.extend(index_to_circuit(i, qubits, base, rng))
        index = mul(i, index)
        ret.extend(int_gate)
        index = mul(int_id, index)

    ret.extend(index_to_circuit(inv(index), qubits, base, rng))

    mapping = {i: q for i, q in enumerate(qubits)}

    # return replace_qubit(ret, qubits)
    return mapping_qubits(ret, mapping)


def RB_two_qubits_IBM(qubits: list[tuple[str]],
                      start: int = 1, max_cycle: int = 50, sweep_points: int = 21,
                      random_times: int = 20,
                      base: Optional[list[list[tuple]]] = None,
                      ini_gate: list = [('X', 'X')], gate_list=[(576, [('Cnot', (0, 1))])],
                      signal: str = 'diag', act_qubits: Optional[list[tuple[str]]] = None, **kw) -> dict:
    """_summary_

    Args:
        qubits (list[tuple[str]]): qubit names.
        start (int, optional): start number. Defaults to 1.
        max_cycle (int, optional): maximum of cycle. Defaults to 500.
        sweep_points (int, optional): sweep points. Defaults to 21.
        random_times (int, optional): random times. Defaults to 20.
        base (Optional[list[list[tuple]]], optional): RB base. Defaults to None.
        ini_gate (list, optional): initial gate. Defaults to [('X', 'X')].
        gate_list (list, optional): insert gates [(num in base, circuit form)]. Defaults to [(576, [('Cnot', (0, 1))])].
        signal (str, optional): signal. Defaults to 'diag'.
        act_qubits (Optional[list[str]], optional): act qubits. Defaults to None.
    """

    act_qubits = qubits if act_qubits is None else act_qubits

    return {
        'init': {
            'name': 'RB Two Qubits IBM',
            'qubits': qubits,
            'signal': signal,
        },
        'sweep_setting': {
            'gate': gate_list,
            'seed': np.random.randint(0xfffffff, size=[random_times]),
            'cycle': generate_log_intlist(max_int=max_cycle, sweep_points=sweep_points, min_int=start),
        },
        'sweep_addition': {
            'circuit': lambda gate, cycle, seed: [
                *[(gate[i], q[i]) for i in range(2)
                  for gate in ini_gate for q in act_qubits],
                *chain.from_iterable(_cr_2RB(qubits=q, cycle=cycle, interleaves=gate,
                                             seed=seed, base=base) for q in act_qubits),
                ('Barrier', tuple(chain.from_iterable(qubits))),
                *[(('Measure', i), q) for i, q in enumerate(chain(*qubits))],
            ],
        },
    }


def QPT_two_qubits(qubits: list[tuple[str]], process: list,
                   signal: str = 'diag', **kw) -> dict:
    """
    [None] Quantum process tomography

    Args:
        qubits (list[tuple[str]]): qubit names.
        process (list): process.
        signal (str, optional): signal. Defaults to 'diag'.
    """

    from waveforms.quantum.tomo import qstOpList, qptInitList
    from waveforms.quantum.rb import mapping_qubits

    return {
        'init': {
            'name': 'QPT',
            'qubits': qubits,
            'signal': signal,
        },
        'sweep_setting': {
            'ini': list(qptInitList(2)),
            'qst': list(qstOpList(2)),
        },
        'sweep_addition': {
            'circuit': lambda ini, qst, process=process: [
                *[(ini[0], q[0]) for q in qubits],
                *[(ini[1], q[1]) for q in qubits],
                *chain.from_iterable(mapping_qubits(process, {ii:qq for ii, qq in enumerate(q)})
                                     for q in qubits),
                *[(qst[i], q[i]) for i in range(2) for q in qubits],
                ('Barrier', tuple(chain.from_iterable(qubits))),
                *[(('Measure', j), q)
                  for j, q in enumerate(chain.from_iterable(qubits))],
            ],
        },
    }



# Author: lizhiyuan  <lizhiyuan@baqis.ac.cn>  2021/01/01
import re


def _parser_header(header):
    pat = re.compile('(.+)\((.*)\)$')
    m = pat.search(header)
    if m:
        operator = (m.group(1))
        qubits = m.group(2).split(',')
    else:
        operator = header
        qubits = []
    return operator, tuple(qubits)


def strindex2stdindex(strindex):
    ''' strindex to stdindex
    Parameters:
        index: str, 
            以operator标志开头；
            如果是多比特门，则用小括号和逗号(xx,xx)表示；
            如果含参，则用下划线连接参数。
            例如，'X', 'U_0_180', 'iSwap(Q1,Q2)', 'CRz(Q1,Q2)_90'
    Return: stdindex,
        (operator, tuple(qubit), tuple(params))
    '''
    header, *params = strindex.split('_')
    operator, qubits = _parser_header(header)
    stdindex = operator, tuple(qubits), tuple(params)
    return stdindex


def tupindex2stdindex(tupindex):
    '''tupindex to stdindex
    与 strindex 相比，将参数独立，避免参数的格式转换
    Parameters:
        index: str or tuple(str,*params), 
            参考 strindex2stdindex，将参数从字符串分离，
            例如，'X', ('U',0,180), 'iSwap(Q1,Q2)', ('CRz(Q1,Q2)',90)
    Return: stdindex,
        (operator, tuple(qubit), tuple(params))
    '''
    if isinstance(tupindex, str):
        header, params = tupindex, tuple()
    else:
        header, params = tupindex
    operator, qubits = _parser_header(header)
    stdindex = operator, tuple(qubits), tuple(params)
    return stdindex


def stdindex2tupindex(stdindex):
    operator, qubits, params = stdindex
    if len(qubits) > 1:
        header = f"{operator}({','.join(qubits)})"
    else:
        header = operator

    if len(params) > 0:
        tupindex = (header, *params)
    else:
        tupindex = header
    return tupindex


def stdindex2strindex(stdindex):
    operator, qubits, params = stdindex
    if len(qubits) > 1:
        header = f"{operator}({','.join(qubits)})"
    else:
        header = operator

    strindex = '_'.join((header, *(str(p) for p in params)))
    return strindex


def index_convert(index, fr='tupindex', to='stdindex'):
    '''index转换的统一接口'''
    if fr != to:
        cov_dict = {
            ('tupindex', 'stdindex'): ('tupindex', 'stdindex'),
            ('stdindex', 'tupindex'): ('stdindex', 'tupindex'),

            ('strindex', 'stdindex'): ('strindex', 'stdindex'),
            ('stdindex', 'strindex'): ('stdindex', 'strindex'),

            ('strindex', 'tupindex'): ('strindex', 'stdindex', 'tupindex'),
            ('tupindex', 'strindex'): ('tupindex', 'stdindex', 'strindex'),
        }
        cov = cov_dict[(fr, to)]
        cov_funcs = (eval(f'{cov[i]}2{cov[i+1]}') for i in range(len(cov)-1))

        for cov_func in cov_funcs:
            index = cov_func(index)
    return index



# Author: lizhiyuan  <lizhiyuan@baqis.ac.cn>  2021/01/01

from qos_tools.core._equal_utils import is_equal
from blinker import Signal
import logging
log = logging.getLogger(__name__)


__all__ = [
    'd2c', 'cons_d2c'
]


class d2c(object):
    '''dict to class

    将字典的调用方式转化为类属性的调用方式；仍支持字典的调用方式；
    字典的关键字须符合类属性命名的规范，可以赋值
    '''

    def __init__(self, d):
        '''传入一个字典或者d2c类的实例，构造一个d2c实例'''
        if isinstance(d, d2c):
            d = d.todict()
        assert isinstance(d, dict), f'not dict! {d}'
        for k, v in d.items():
            _v = d2c(v) if isinstance(v, dict) else v
            self.__dict__.update({k: _v})

    def __getitem__(self, key):
        return self.get(key)

    def __setitem__(self, key, value):
        return self.set(key, value)

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()

    def items(self):
        return self.__dict__.items()

    def todict(self):
        '''从d2c实例提取出原字典'''
        d = {}
        for k, v in self.__dict__.items():
            _v = v.todict() if isinstance(v, d2c) else v
            d.update({k: _v})
        return d

    def get(self, keys, default=None, splitsymbol='.'):
        '''根据关键词连接的字符串或者列表，获取对应的值

        Parameters:
            keys: 使用符号串联key的字符串，比如'pulse.width.value';
                或者key顺序列表，比如['pulse','width','value']
            default: keys不存在时返回一个默认值
            splitsymbol: 字符串的分割符号，默认为 '.'
        Return:
            返回键值串对应的值
        '''
        try:
            keys = keys.split(splitsymbol) if isinstance(keys, str) else keys
            value = self
            for k in keys:
                value = getattr(value, k)
            ##
            # value=getattr(self,keys.pop(0))
            # if keys:
            #     value=value.get(keys) #递归
        except AttributeError:
            value = default
        return value

    def set(self, keys, value, check=False, splitsymbol='.'):
        '''根据关键词连接的字符串或者列表，设置对应的值

        Parameters:
            keys: 使用符号串联key的字符串，比如'pulse.width.value';
                或者key顺序列表，比如['pulse','width','value']
            value: 待设入的值
            check: bool, 是否检查设入值与原值相等
            splitsymbol: 字符串的分割符号，默认为 '.'
        Return:
            bool, True 表示value改变并且已设置，False 表示未变
        '''
        if check:
            if is_equal(value, self.get(keys)):
                return False
        keys = keys.split(splitsymbol) if isinstance(keys, str) else keys
        ins = self
        for k in keys[:-1]:
            ins = getattr(ins, k)
        assert hasattr(ins, keys[-1]), f'关键词 {keys[-1]} 不存在!'
        setattr(ins, keys[-1], value)
        ##
        # if len(keys)>1:
        #     getattr(self,keys.pop(0)).set(keys,value) ##递归
        # else:
        #     assert hasattr(self,keys[0]) ##必须是字典中已有的关键词
        #     setattr(self,keys[0],value)
        return True


class cons_d2c(d2c):
    '''包含约束的d2c类'''

    def __init__(self, d, constrains=[], init=True):
        '''
        Parameters:
            d: 一个字典
            constrains: 约束条目，每个元素都为一个三元元组（deps, func, target），例如，
                [
                    (('k1.k11','k1.k12'),   (lambda v1,v2:v1+v2),   'k2'       ),
                    (('k3','k4'),           (lambda v1,v2:(v1,v2)), ('k5','k6')),
                ]
            init: bool, 是否使用约束条件初始化传入的字典
        '''
        self.__d2c_instance = d if isinstance(d, d2c) else d2c(d)
        self.sig = Signal()
        self.constrains = constrains
        self.subscribers = self.__init_cons(constrains, init=init)

    def __init_cons(self, constrains, init=True):
        subscribers = []
        for deps, func, target in constrains:
            subs = _subscrib(deps, func, target, self)
            subscribers.append(subs)
            for _k in deps:
                self.sig.connect(subs, sender=_k)
        if init:
            deps_set = set()
            for deps, func, target in constrains:
                deps_set.update(deps)
            for dep in deps_set:
                self.sig.send(dep)
        return subscribers

    def __getattr__(self, item):
        return self.__d2c_instance.get(item)

    # def __setattr__(self,item,value):  # 有bug
    #     return self.__d2c_instance.set(item,value)

    def todict(self):
        return self.__d2c_instance.todict()

    def set(self, k, v, check=True):
        _changed = self.__d2c_instance.set(k, v, check=check)
        if _changed:
            self.sig.send(k)
        return _changed

    def get(self, k, default=None):
        return self.__d2c_instance.get(k, default=default)

    def update(self, d, check=True):
        k_changed = []
        for k, v in d.items():
            _changed = self.__d2c_instance.set(k, v, check=check)
            if _changed:
                k_changed.append(k)
        for _k in k_changed:
            self.sig.send(_k)


def _subscrib(deps, func, target, cons_d2c_ins):
    '''根据需要的参量产生一个订阅函数
    Parameters:
        deps,func,target: 分别为 依赖的key[列表或元组]、约束函数、目标[单key(字符串)，或者多key(列表/元组)]，
            比如 (('k1','k2'), lambda v1,v2:v1+v2, 'k3'), func的返回值应与target数目一致
        cons_d2c_ins: 上面cons_d2c的一个实例
    Return:
        返回一个订阅函数
    '''
    if isinstance(target, str):
        def subscriber(send):
            arg = [cons_d2c_ins.get(k_dep) for k_dep in deps]
            if None not in arg:
                value = func(*arg)
                cons_d2c_ins.set(target, value)
            else:  # 这里只打出警告：一些依赖变量的值为None，打印出所有的键值对
                log.warning(
                    'WARNING: ' + ', '.join([f'{k} is {v}' for k, v in zip(deps, arg)]))
    else:
        def subscriber(send):
            arg = [cons_d2c_ins.get(k_dep) for k_dep in deps]
            if None not in arg:
                value = func(*arg)
                d = dict(zip(target, value))
                cons_d2c_ins.update(d)
            else:  # 这里只打出警告：一些依赖变量的值为None
                log.warning(
                    'WARNING: ' + ', '.join([f'{k} is {v}' for k, v in zip(deps, arg)]))
    return subscriber



# Rebuild: Liu Pei  <liupei200546@163.com>  2022/05/04

from itertools import chain
from typing import Optional, Union

import numpy as np

__all__ = [
    'RamseyLike', 'T1'
]


def RamseyLike(qubits: list[str],
               rotate: float, time_length: float, points_in_one_period: int = 10,
               name: str = 'Ramsey',
               signal: str = 'iq_avg', act_qubits: Optional[list[str]] = None, **kw) -> dict:
    """
    ['time'] SpinEcho, Ramsey, XY16, XY8, XY4, CP, CPMG and UDD circuit.

    Args:
        qubits (list[str]): the qubit name.
        rotate (float): offset of frequency.
        time_length (float): length of delay time.
        points_in_one_period (int, optional): points per period. Defaults to 10.
        signal (str, optional): signal. Defaults to 'iq_avg'.
        act_qubits (Optional[list[str]], optional): act qubits. Defaults to None.
        n(int): parameter for certain circuit via `kw`

    Returns:
        dict: config
    """

    assert name in ['SpinEcho', 'Ramsey', 'XY16', 'XY8', 'XY4',
                    'CP', 'CPMG', 'UDD', ], 'Invalid circuit type'
    kwn = {}
    if name in ['CP', 'CPMG', 'UDD']:
        kwn['n'] = kw.get('n', 1)

    act_qubits = qubits if act_qubits is None else act_qubits
    import waveforms.quantum as utils

    return {
        'init': {
            'name': name,
            'qubits': qubits,
            'signal': signal,
        },
        'sweep_setting': {
            'time': np.linspace(0, time_length, round(time_length*rotate*points_in_one_period)+1),
        },
        'sweep_addition': {
            'circuit': lambda time: [
                *chain.from_iterable(getattr(utils, name)(qubit=q,
                                     t=time, f=rotate, **kwn) for q in act_qubits),
                ('Barrier', tuple(qubits)),
                *[(('Measure', j), q) for j, q in enumerate(qubits)],
            ],
        },
    }


def T1(qubits: list[str],
       time_length: Union[float, list], mode: str = 'linear', sweep_points: int = 51,
       signal: str = 'iq_avg', act_qubits: Optional[list[str]] = None, **kw) -> dict:
    """
    [f'{q}'] T1.

    Args:
        qubits (list[str]): the qubit name.
        time_length (Union[float, list]): the delay time length, for each qubit(in the type list of float) or for all qubits(int the type float).
        sweep_points (int, optional): sweep points. Defaults to 51.
        mode (str, optional): in ['linear' (for linear time sampling), 'log' (for log time sampling)]. Defaults to 'linear'.
        signal (str, optional): signal. Defaults to 'iq_avg'.
        act_qubits (Optional[list[str]], optional): act qubits. Defaults to None.
        signal (str, optional): signal type. Defaults to 'raw'.

    Raises:
        ValueError: The time_length is illegal.
        ValueError: [description]

    Returns:
        dict: config.
    """

    act_qubits = qubits if act_qubits is None else act_qubits

    if isinstance(time_length, list) and len(time_length) != len(qubits):
        raise ValueError('The time_length is illegal')
    elif isinstance(time_length, float):
        time_length = [time_length for _ in qubits]
    else:
        raise ValueError('The time_length is illegal')

    timeList = []
    if mode in ['log']:
        for j, _ in enumerate(qubits):
            timeList.append(np.logspace(
                1e-9, np.log10(time_length), sweep_points))
    else:
        for j, _ in enumerate(qubits):
            timeList.append(np.concatenate((np.linspace(0, time_length[j]/3, sweep_points//3*2),
                            np.linspace(time_length[j]/3, time_length[j], sweep_points-sweep_points//3*2+1)[1:])))

    return {
        'init': {
            'name': 'T1',
            'qubits': qubits,
            'signal': signal,
        },
        'sweep_setting': {
            tuple(act_qubits): tuple(timeList),
        },
        'sweep_addition': {
            'circuit': lambda **kw: [
                *[(('rfUnitary', np.pi, 0), q) for q in act_qubits],
                *[(('Delay', kw[q]), q) for q in act_qubits],
                ('Barrier', tuple(qubits)),
                *[(('Measure', j), q) for j, q in enumerate(qubits)],
            ],
        },
    }



# Rebuild: Liu Pei  <liupei200546@163.com>  2022/02/25

from typing import Optional, Union

import kernel
import numpy as np

from .tools import generate_spanlist

__all__ = [
    'S21_change_lo', 'S21_change_awg',
    'Readout_frequency', 'Readout_amp', 'Scatter',
]


def S21_change_lo(qubits: list[str],
                  center: Optional[Union[float, list[float]]] = None, delta: Optional[float] = None, st: Optional[float] = None, ed: Optional[float] = None, sweep_points: int = 101, mode: str = 'linear',
                  signal: str = 'iq_avg', **kw) -> dict:
    """
    [f'{q}'] Measure S21 without constraints, change local frequency.

    Args:
        qubits (list[str]): qubit names.
        center (Optional[Union[float, list[float]]], optional): sweep center. Defaults to None.
        delta (Optional[float], optional): sweep span. Defaults to None.
        st (Optional[float], optional): sweep start. Defaults to None.
        ed (Optional[float], optional): sweep end. Defaults to None.
        sweep_points (int, optional): sweep points. Defaults to 101.
        mode (str, optional): sweep mode. Defaults to 'linear'.
        signal (str, optional): signal. Defaults to 'iq_avg'.
    """

    if center is None:
        center = [kernel.get(
            f'gate.Measure.{q}.params.frequency') for q in qubits]
    elif isinstance(center, float):
        center = [center]*len(qubits)

    sweep_list = generate_spanlist(
        center=0, delta=delta, st=st, ed=ed, sweep_points=sweep_points, mode=mode)
    measures = set(kernel.get(f'{q}.probe') for q in qubits)

    return {
        'init': {
            'name': 'S21',
            'qubits': qubits,
            'signal': signal,
        },
        'setting': {
            'compile_once': True,
            'circuit': [(('Measure', j), q) for j, q in enumerate(qubits)],
        },
        'sweep_config':
        dict(
            list({
                q: {
                    'addr': f'gate.Measure.{q}.params.frequency',
                }
                for q in qubits
            }.items()) +
            list({m: {
                'addr': f'{m}.setting.LO',
            }
                for m in measures}.items())),
        'sweep_setting': {
            tuple([
                *qubits,
                *measures,
            ]): tuple([
                *[
                    sweep_list + center[j]
                    for j, _ in enumerate(qubits)
                ],
                *[
                    sweep_list + kernel.get(f'{m}.setting.LO')
                    for m in measures
                ],
            ]),
        },
    }


def S21_change_awg(qubits: list[str],
                   center: Optional[Union[float, list[float]]] = None, delta: Optional[float] = None, st: Optional[float] = None, ed: Optional[float] = None, mode: str = 'linear', sweep_points: int = 101,
                   signal: str = 'iq_avg', **kw) -> dict:
    """
    [f'Q{i}'] Measure S21 without constraints, change awg frequency.

    Args:
        qubits (list[str]): qubit names.
        center (Optional[Union[float, list[float]]], optional): sweep center. Defaults to None.
        delta (Optional[float], optional): sweep span. Defaults to None.
        st (Optional[float], optional): sweep start. Defaults to None.
        ed (Optional[float], optional): sweep end. Defaults to None.
        sweep_points (int, optional): sweep points. Defaults to 101.
        mode (str, optional): sweep mode. Defaults to 'linear'.
        signal (str, optional): signal. Defaults to 'iq_avg'.
    """

    if center is None:
        center = [kernel.get(
            f'gate.Measure.{q}.params.frequency') for q in qubits]
    elif isinstance(center, float):
        center = [center]*len(qubits)

    sweep_list = generate_spanlist(
        center=0, delta=delta, st=st, ed=ed, sweep_points=sweep_points, mode=mode)

    return {
        'init': {
            'name': 'S21',
            'qubits': qubits,
            'signal': signal,
        },
        'setting': {
            'circuit':
            [(('Measure', j), q) for j, q in enumerate(qubits)],
        },
        'sweep_config': {
            q: {
                'addr': f'gate.Measure.{q}.params.frequency',
            }
            for q in qubits
        },
        'sweep_setting': {
            tuple(qubits): tuple([
                sweep_list + center[j]
                for j, i in enumerate(qubits)
            ]),
        },
    }


def Readout_frequency_lo(qubits: list[str],
                         center: Optional[Union[float, list[float]]] = None, delta: Optional[float] = None, st: Optional[float] = None, ed: Optional[float] = None, mode: str = 'linear', sweep_points: int = 101,
                         gate: list = [['I'], [('rfUnitary', np.pi, 0)]],
                         signal: str = 'iq', **kw) -> dict:
    """
    [f'{q}'] Read frequency measurement under 0 state and 1 state.

    Args:
        qubits (list[str]): qubit names.
        center (Optional[Union[float, list[float]]], optional): sweep center. Defaults to None.
        delta (Optional[float], optional): sweep span. Defaults to None.
        st (Optional[float], optional): sweep start. Defaults to None.
        ed (Optional[float], optional): sweep end. Defaults to None.
        sweep_points (int, optional): sweep points. Defaults to 101.
        mode (str, optional): sweep mode. Defaults to 'linear'.
        gate (list): initial gate. Defaults to ['I', ('rfUnitary', np.pi, 0)].
        signal (str, optional): signal. Defaults to 'iq'.
    """

    if center is None:
        center = [kernel.get(
            f'gate.Measure.{q}.params.frequency') for q in qubits]
    elif isinstance(center, float):
        center = [center]*len(qubits)

    sweep_list = generate_spanlist(
        center=0, delta=delta, st=st, ed=ed, sweep_points=sweep_points, mode=mode)
    measures = set(kernel.get(f'{q}.probe') for q in qubits)

    return {
        'init': {
            'name': 'Readout Frequency',
            'qubits': qubits,
            'signal': signal,
        },
        'sweep_config': dict(
            list({
                q: {
                    'addr': f'gate.Measure.{q}.params.frequency',
                } for q in qubits
            }.items())+list({
                m: {
                    'addr': f'{m}.setting.LO',
                } for m in measures
            }.items())
        ),
        'sweep_setting': {
            ('gate', 'skip'): tuple([gate, list(range(1, 1+len(gate)))]),
            tuple([
                *qubits,
                *measures,
            ]): tuple([
                *[sweep_list+center[j] for j, _ in enumerate(qubits)],
                *[sweep_list+kernel.get(f'{m}.setting.LO') for m in measures],
            ]),
        },
        'sweep_addition': {
            'circuit': lambda gate: [
                *[(gat, q) for gat in gate for q in qubits],
                ('Barrier', tuple(qubits)),
                *[(('Measure', j), q) for j, q in enumerate(qubits)],
            ],
        },
    }


def Readout_frequency_awg(qubits: list[str],
                          center: Optional[Union[float, list[float]]] = None, delta: Optional[float] = None, st: Optional[float] = None, ed: Optional[float] = None, mode: str = 'linear', sweep_points: int = 101,
                          gate: list = [['I'], [('rfUnitary', np.pi, 0)]],
                          signal: str = 'iq', **kw) -> dict:
    """
    [f'{q}'] Read frequency measurement under 0 state and 1 state.

    Args:
        qubits (list[str]): qubit names.
        center (Optional[Union[float, list[float]]], optional): sweep center. Defaults to None.
        delta (Optional[float], optional): sweep span. Defaults to None.
        st (Optional[float], optional): sweep start. Defaults to None.
        ed (Optional[float], optional): sweep end. Defaults to None.
        sweep_points (int, optional): sweep points. Defaults to 101.
        mode (str, optional): sweep mode. Defaults to 'linear'.
        gate (list): initial gate. Defaults to ['I', ('rfUnitary', np.pi, 0)].
        signal (str, optional): signal. Defaults to 'iq'.
    """

    if center is None:
        center = [kernel.get(
            f'gate.Measure.{q}.params.frequency') for q in qubits]
    elif isinstance(center, float):
        center = [center]*len(qubits)

    sweep_list = generate_spanlist(
        center=0, delta=delta, st=st, ed=ed, sweep_points=sweep_points, mode=mode)

    return {
        'init': {
            'name': 'Readout Frequency',
            'qubits': qubits,
            'signal': signal,
        },
        'sweep_setting': {
            'gate': gate,
            tuple(qubits): tuple([
                *[sweep_list+center[j] for j, _ in enumerate(qubits)],
            ]),
        },
        'sweep_addition': {
            'circuit': lambda gate, **kw: [
                *[(gat, q) for gat in gate for q in qubits],
                ('Barrier', tuple(qubits)),
                *[(('Measure', j, ('with', ('param:frequency', kw[q]))), q)
                  for j, q in enumerate(qubits)],
            ],
        },
    }


def Readout_amp(qubits: list[str],
                center: Optional[float] = None, st: Optional[float] = None, ed: Optional[float] = None, mode: str = 'linear', sweep_points: int = 101,
                gate: list = [['I'], [('rfUnitary', np.pi, 0)]],
                signal: str = 'iq', **kw) -> dict:
    """
    [f'{q}'] Read amp measurement under 0 state and 1 state. 

    Args:
        qubits (list[str]): qubit names.
        center (Optional[Union[float, list[float]]], optional): sweep center. Defaults to None.
        st (Optional[float], optional): sweep start. Defaults to None.
        ed (Optional[float], optional): sweep end. Defaults to None.
        sweep_points (int, optional): sweep points. Defaults to 101.
        mode (str, optional): sweep mode. Defaults to 'linear'.
        gate (list): initial gate. Defaults to ['I', ('rfUnitary', np.pi, 0)].
        signal (str, optional): signal. Defaults to 'iq'.
    """

    span_list = generate_spanlist(
        center=center, st=st, ed=ed, mode=mode, sweep_points=sweep_points)

    return {
        'init': {
            'name': 'Readout Amp',
            'qubits': qubits,
            'signal': signal,
        },
        'sweep_setting': {
            'gate': gate,
            tuple(qubits): tuple(span_list*kernel.get(f'gate.Measure.{q}.params.amp') for q in qubits),
        },
        'sweep_addition': {
            'circuit': lambda gate, **kw: [
                *[(gat, q) for gat in gate for q in qubits],
                ('Barrier', tuple(qubits)),
                *[(('Measure', j, ('with', ('param:frequency', kw[q]))), q)
                  for j, q in enumerate(qubits)],
            ],
        },
    }


def Scatter(qubits: list[str],
            repeat: int = 1, gate: list = [['I'], [('rfUnitary', np.pi, 0)]],
            signal: str = 'iq', **kw) -> dict:
    """
    [None] 0 state and 1 state classification.

    Args:
        qubits (list[str]): qubit names.
        repeat (int, optional): repeat times. Defaults to 1.
        gate (list, optional): applied gate. Defaults to ['I', ('rfUnitary', np.pi, 0)].
        signal (str, optional): signal. Defaults to 'iq'.
    """

    return {
        'init': {
            'name': 'Scatter',
            'qubits': qubits,
            'signal': signal,
        },
        'sweep_setting': {
            'gate': gate,
            'repeat': np.arange(repeat),
        },
        'sweep_addition': {
            'circuit': lambda gate: [
                *[(gat, q) for gat in gate for q in qubits],
                ('Barrier', tuple(qubits)),
                *[(('Measure', j), q) for j, q in enumerate(qubits)],
            ],
        },
    }


def ReadoutDelay(qubits: list[str], cavity_frequencies: dict, delay_list: list[tuple], intrinsic_delay: float,
                 pulse_time: float = 100e-9, pulse_amp: float = 1,
                 signal: float = 'population', **kw) -> dict:
    """
    ['delay'] Readout time delay.

    Args:
        qubits (list[str]): qubit names.
        cavity_frequencies (dict): cavity frequencies.
        delay_list (list[tuple]): [(st, ed, steps) pair].
        intrinsic_delay (float): intrinsic delay.
        pulse_time (float, optional): pulse duration. Defaults to 100e-9.
        pulse_amp_times (float, optional): pulse amp. Defaults to 1.
        signal (float, optional): signal. Defaults to 'population'.
    """                 
    
    from waveforms.waveform import square, mixing
    
    readout_pulse = {}
    for q in qubits:
        readout_pulse[q], _ = mixing(pulse_amp*square(pulse_time)>>(pulse_time/2), freq=cavity_frequencies[q])
        
    return {
        'init': {
            'name': 'Readout Delay',
            'qubits': qubits,
            'signal': signal,
        },
        'sweep_setting': {
            'delay': np.concatenate([np.arange(st, ed, step) for st, ed, step in delay_list]),
        },
        'sweep_addition': {
            'circuit': lambda delay: [
                *[(('Pulse', 'readoutLine.RF', readout_pulse[q]), q) for q in qubits],
                *[(('Delay', delay+intrinsic_delay), q) for q in qubits],
                *[(('rfUnitary', np.pi/2, 0), q) for q in qubits],
                *[(('Measure', j), q) for j, q in enumerate(qubits)],
            ],
        },
    }



# Author: lizhiyuan  <lizhiyuan@baqis.ac.cn>  2021/08/08

def qlispcmd2stdcmd(qlispcmd):
    '''
    qlispcmd: 
        ((operator, *params), qubits)
        (operator, qubits)
        (operator, qubit)
    stdcmd: 
        (operator,qubits,params)
    '''
    op_para, qubits = qlispcmd
    if isinstance(op_para, tuple):
        operator, *params = op_para
    else:
        operator = op_para
        params = tuple()
    if isinstance(qubits, str):
        qubits = (qubits,)
    stdcmd = (operator, qubits, params)
    return stdcmd


def stdcmd2qlispcmd(stdcmd):
    '''
    qlispcmd:
        ((operator, *params), qubits)
        (operator, qubits)
        (operator, qubit)

    stdcmd:
        (operator,qubits,params)
    '''

    operator, qubits, params = stdcmd
    if len(qubits) == 1:
        qubits, = qubits

    if params:
        qlispcmd = ((operator, *params), qubits)
    else:
        qlispcmd = (operator, qubits)
    return qlispcmd


def cmd_convert(cmd, fr='qlispcmd', to='stdcmd'):
    '''index转换的统一接口'''
    if fr != to:
        cov_dict = {
            ('qlispcmd', 'stdcmd'): ('qlispcmd', 'stdcmd'),
            ('stdcmd', 'qlispcmd'): ('stdcmd', 'qlispcmd'),

        }
        cov = cov_dict[(fr, to)]
        cov_funcs = (eval(f'{cov[i]}2{cov[i+1]}') for i in range(len(cov)-1))

        for cov_func in cov_funcs:
            cmd = cov_func(cmd)
    return cmd



# Rebuild & Author: Liu Pei <liupei200546@163.com>  2021/08/14


from typing import Union

import numpy as np
from qos_tools.analyzer.tools import get_rotate_phi


def generate_1D_mean(result: dict, index: list[str], shape: list, sig: str, state_limit: int) -> list[dict]:
    assert sig in ['data', 'states']
    for j, i in enumerate(index):
        ret = {
            'x': np.asarray(result['index'][index[j]][:shape[0]]),
            'y': np.mean(result[sig][:, :, j], axis=-1).reshape(shape),
        }
        if sig=='states':
            count = result['states'][:, :, j]
            population = []
            for state in range(state_limit):
                population.append(np.mean(np.where(count==state, 1, 0), axis=-1).reshape(shape))
            ret['population'] = np.asarray(population)
            ret['y'] = population[1]
        yield ret


def format_conversion(scanner: str, result: dict, qubits: list[Union[int, tuple]], signal: str = 'raw', state_limit: int = 2) -> list[dict]:

    sig = 'states' if signal == 'state' else 'data'

    if scanner in [
        'S21_base_change_lo_with_constrains'
    ]:
        ret = []
        for item in generate_1D_mean(state_limit=state_limit, result=result, index=['delta' for _ in qubits], shape=[result[sig].shape[0], ], sig=sig):
            item['y'] = np.abs(item['y'])
            ret.append(item)
        return ret

    elif scanner in [
        'S21_base_change_lo_without_constrains',
        'S21_base_change_awg',
        'S21_change_lo_without_constrains',
    ]:
        ret = []
        for item in generate_1D_mean(state_limit=state_limit, result=result, index=[f'Q{i}' for i in qubits], shape=[result[sig].shape[0], ], sig=sig):
            item['y'] = np.abs(item['y'])
            ret.append(item)
        return ret

    elif scanner in [
        'Spectrum_base_change_awg_without_constrains',
        'Spectrum_base_change_lo_without_constrains',
    ]:
        ret = []
        for item in generate_1D_mean(state_limit=state_limit, result=result, index=[f'Q{i}' for i in qubits], shape=[result[sig].shape[0], ], sig=sig):
            item['y'] = np.abs(get_rotate_phi(item['y']-np.mean(item['y'])))
            ret.append(item)
        return ret

    elif scanner in [
        'PowerRabi_base_scaleTimes',
        'PowerRabi_base_n_pulse',
        'TimeRabi_base_durationTimes',
        'XXDelta_without_constrain'
    ]:
        ret = []
        for item in generate_1D_mean(state_limit=state_limit, result=result, index=[f'Q{i}' for i in qubits], shape=[result[sig].shape[0], ], sig=sig):
            if signal == 'raw':
                item['y'] = np.real(get_rotate_phi(item['y']-item['y'][0]))
            ret.append(item)
        return ret

    elif scanner in [
        'T1_without_constrains'
    ]:
        ret = []
        for item in generate_1D_mean(state_limit=state_limit, result=result, index=[f'Q{i}' for i in qubits], shape=[result[sig].shape[0], ], sig=sig):
            if signal == 'raw':
                item['y'] = np.real(get_rotate_phi(item['y']-item['y'][-1]))
            ret.append(item)
        return ret

    elif scanner in [
        'Ramsey_without_constrains',
        'SpinEcho_without_constrains',
        'CPMG_without_constrains',
        'CP_without_constrains',
        'ReadoutDelay_without_constrains',
        'RTO_without_constrains'
    ]:
        ret = []
        for item in generate_1D_mean(state_limit=state_limit, result=result, index=['time' for _ in qubits], shape=[result[sig].shape[0], ], sig=sig):
            if signal == 'raw':
                item['y'] = np.real(get_rotate_phi(item['y']-item['y'][0]))
            ret.append(item)
        return ret

    elif scanner in [
        'ReadoutFrequency_without_constrains',
        'ReadoutAmp_without_constrains',
    ]:
        ret = []
        lth = (result['data'].shape[0])//state_limit
        for j, i in enumerate(qubits):
            ret.append({
                'x': np.asarray(result['index'][f'Q{i}'])[:lth],
            })
            for cur_state in range(state_limit):
                ret[-1][f'S{cur_state}'] = result[sig][cur_state*lth:cur_state*lth+lth, :, j]
        return ret

    elif scanner in [
        'Scatter2',
        'Scatter3',
    ]:
        ret = []
        for j, i in enumerate(qubits):
            ret.append({})
            for cur_state in range(state_limit):
                ret[-1][f'S{cur_state}'] = result[sig][cur_state, :, j]
        return ret

    elif scanner in [
        'AllXY_without_constrains',
        'DRAGCheck_without_constrain'
    ]:
        ret = []
        for j, i in enumerate(qubits):
            item = {'y': np.mean(result[sig][:, :, j], axis=-1)}
            if signal == 'raw':
                item['y'] = np.real(get_rotate_phi(item['y']-item['y'][0]))
            else:
                count = result['states'][:, :, j]
                population = []
                for state in range(state_limit):
                    population.append(np.mean(np.where(count==state, 1, 0), axis=-1))
                item['population'] = np.asarray(population)
                item['y'] = population[1]
            ret.append(item)
        return ret
    elif scanner in [
        'AllXY6and16_alpha_without_constrains',
        'AllXY7and8_delta_without_constrains',
        'AllXY11and12_beta_without_constrains',
    ]:
        ret = []
        for j, i in enumerate(qubits):
            x = np.asarray(result['index'][f'Q{i}'])
            y = np.mean(result[sig][:, :, j], axis=-1)
            ret.append({
                'x': x[:y.shape[0]//2],
            })
            if signal == 'raw':
                y = np.real(get_rotate_phi(y))
            else:
                count = result['states'][:, :, j]
                population = []
                for state in range(state_limit):
                    population.append(np.mean(np.where(count==state, 1, 0), axis=-1).reshape([y.shape[0]//2, 2]))
                ret[-1]['population'] = np.asarray(population)
                ret[-1]['y'] = population[1]
            ret[-1]['y1'] = y[:y.shape[0]//2],
            ret[-1]['y2'] = y[y.shape[0]//2:],
        return ret
    elif scanner in [
        'PiErrorOscillation_without_constrain',
    ]:
        ret = []
        for item in generate_1D_mean(state_limit=state_limit, result=result, index=['count' for _ in qubits], shape=[result[sig].shape[0], ], sig=sig):
            if signal == 'raw':
                item['y'] = np.real(get_rotate_phi(item['y']-item['y'][0]))
            ret.append(item)
        return ret

    elif scanner in [
        'RB_single_qubit_without_constrain'
    ]:
        ret = []
        for j, i in enumerate(qubits):
            item = {'y': np.mean(result[sig][:, :, j], axis=-1)}
            if signal == 'raw':
                item['y'] = np.real(get_rotate_phi(item['y']))
            else:
                count = result['states'][:, :, j]
                population = []
                for state in range(state_limit):
                    population.append(np.mean(np.where(count==state, 1, 0), axis=-1))
                item['population'] = np.asarray(population)
                item['y'] = population[1]
            ret.append(item)
        return ret

    elif scanner in [
        'XEB_single_qubit_without_constrain'
    ]:
        assert signal == 'count', 'signal type is invalid.'

        cycle = list(set(result['index']['cycle']))
        seeds = list(set(result['index']['random_times']))

        return {
            'cycle': cycle,
            'seeds': seeds,
            'counts': result['counts'],
        }

    elif scanner in [
        'nop',
    ]:
        return [{} for _ in qubits]
    else:
        raise ValueError('scanner is not supported')



# Author: Liu Pei  <liupei200546@163.com>  2021/08/14

from typing import Any, Literal, Optional


def create_dev(driver: Optional[str] = None, addr: Optional[str] = None, type: Literal['driver', 'remote'] = 'driver', srate: float = -1, **kw) -> dict:
    """
    Quick method for create a dev isinstance.

    Args:
        driver (Optional[str], optional): driver name. Defaults to None.
        addr (Optional[str], optional): address. Defaults to None.
        type (Literal[, optional): backend type. Defaults to 'driver'.
        srate (float, optional): sample rate. Defaults to -1.

    Returns:
        dict: keys to be completed.
    """
    return {
        "addr": addr,
        "name": driver,
        "type": type,
        "srate": srate,
    }


def add_setting(setting_LO: Optional[None] = None, setting_POW: Optional[float] = None, setting_OFFSET: Optional[float] = None, **kw) -> dict:
    """
    Quick method for add a `Setting` session.

    Args:
        setting_LO (Optional[None], optional): local frequency. Defaults to None.
        setting_POW (Optional[float], optional): power. Defaults to None.
        setting_OFFSET (Optional[float], optional): offset. Defaults to None.

    Returns:
        dict: keys to be completed.
    """
    return {
        "LO": setting_LO,
        "POW": setting_POW,
        "OFFSET": setting_OFFSET,  # 基本不用
    }


def add_waveform(waveform_SR: Optional[float] = None, waveform_LEN: Optional[float] = None, waveform_DDS_LO: Optional[float] = None, **kw) -> dict:
    """
    Quick method for add a `waveform` session.

    Args:
        waveform_SR (Optional[float], optional): sample rate for waveform. Defaults to None.
        waveform_LEN (Optional[float], optional): length for waveform. Defaults to None.
        waveform_DDS_LO (Optional[float], optional): DDS or Tek70000 local frequency. Defaults to None.

    Returns:
        dict: keys to be completed.
    """
    return {
        "SR": waveform_SR,
        "LEN": waveform_LEN,
        "DDS_LO": waveform_DDS_LO,
        "SW": "zero()",
        "TRIG": "zero()",
        "RF": "zero()",
        "Z": "zero()",
    }


def add_channel(channel_I: Optional[str] = None, channel_Q: Optional[str] = None, channel_LO: Optional[str] = None,
                channel_DDS: Optional[str] = None, channel_SW: Optional[str] = None, channel_TRIG: Optional[str] = None,
                channel_Z: Optional[str] = None, **kw) -> dict:
    """
    Quick method for add a `channel` session.

    Args:
        channel_I (Optional[str], optional): I channel equipment. Defaults to None.
        channel_Q (Optional[str], optional): Q channel equipment. Defaults to None.
        channel_LO (Optional[str], optional): local channel equipment. Defaults to None.
        channel_DDS (Optional[str], optional): DDS or Tek70000 channel equipment. Defaults to None.
        channel_SW (Optional[str], optional): swich channel equipment. Defaults to None.
        channel_TRIG (Optional[str], optional): trig channel equipment. Defaults to None.
        channel_Z (Optional[str], optional): Z channel equipment. Defaults to None.

    Returns:
        dict: keys to be completed.
    """
    return {
        "I": channel_I,
        "Q": channel_Q,
        "LO": channel_LO,
        "DDS": channel_DDS,
        "SW": channel_SW,
        "TRIG": channel_TRIG,
        "Z": channel_Z,
    }


def add_calibration_channel(delay: float = 0, distortion: Any = None) -> dict:
    """
    Quick method for add a `calibration` session.

    Args:
        delay (float, optional): delay time. Defaults to 0.
        distortion (Any, optional): method for distortion. Defaults to 0.

    Returns:
        dict: keys to be completed.
    """
    return {
        "delay": delay,
        "distortion": distortion,
    }


def add_calibration(calibration_I_delay: float = 0, calibration_I_distortion: Any = None, calibration_Q_delay: float = 0, calibration_Q_distortion: Any = None,
                    calibration_Z_delay: float = 0, calibration_Z_distortion: Any = None, calibration_DDS_delay: float = 0, calibration_DDS_distortion: Any = None,
                    calibration_TRIG_delay: float = 0, calibration_TRIG_distortion: Any = None, **kw) -> dict:
    """
    Quick method for add a `calibration` instance.

    Args:
        ref to `add_calibration_channel`.

    Returns:
        dict: keys to be completed.
    """
    return {
        "I": add_calibration_channel(calibration_I_delay, calibration_I_distortion),
        "Q": add_calibration_channel(calibration_Q_delay, calibration_Q_distortion),
        "Z": add_calibration_channel(calibration_Z_delay, calibration_Z_distortion),
        "DDS": add_calibration_channel(calibration_DDS_delay, calibration_DDS_distortion),
        "TRIG": add_calibration_channel(calibration_TRIG_delay, calibration_TRIG_distortion)
    }


def create_qubit(probe: Optional[str] = None, couplers: list[str] = [], **kw) -> dict:
    """
    Quick method for create a qubit isinstance.

    Args:
        probe (Optional[str], optional): probe name. Defaults to None.
        couplers (list[str], optional): couplers. Defaults to [].

    Returns:
        dict: keys to be completed.
    """
    return {
        "probe": probe,
        "couplers": couplers,
        "setting": add_setting(**kw),
        "waveform": add_waveform(**kw),
        "channel": add_channel(**kw),
        "calibration": add_calibration(**kw),
    }


def create_coupler(qubits: list[str] = [], **kw) -> dict:
    """
    Quick method for create a coupler isinstance.

    Args:
        qubits (list[str], optional): qubit list. Defaults to [].

    Returns:
        dict: keys to be completed.
    """
    return {
        "qubits": qubits,
        "setting": add_setting(**kw),
        "waveform": add_waveform(**kw),
        "channel": add_channel(**kw),
        "calibration": add_calibration(**kw),
    }


def create_probe(qubits: list[str] = [], adcsr: Optional[float] = None, channel_ADC: Optional[str] = None, pointNum: int = 2048, **kw) -> dict:
    """
    Quick method for create a probe isinstance.

    Args:
        qubits (list[str], optional): qubit list. Defaults to [].
        adcsr (Optional[float], optional): sample rate for adc. Defaults to None.
        channel_ADC (Optional[str], optional): ADC channel. Defaults to None.
        setting_SHOT (int, optional): repeat shots. Defaults to 1024.

    Returns:
        dict: keys to be completed.
    """

    setting = add_setting(**kw)
    setting.pop('OFFSET')
    setting['SHOT'] = 1024
    setting['TRIGD'] = 0
    setting['PNT'] = pointNum

    channel = add_channel(**kw)
    channel["ADC"] = channel_ADC

    return {
        "qubits": qubits,
        "adcsr": adcsr,
        "setting": setting,
        "waveform": add_waveform(**kw),
        "channel": channel,
        "calibration": add_calibration(**kw),
    }


def create_gate_measure(measure_frequency: Optional[float] = None, measure_duration: Optional[float] = None, measure_amp: Optional[float] = None,
                        measure_phi: float = 0, measure_threshold: float = 0, measure_type: str = "default", measure_bias: float = 0, **kw) -> dict:
    """
    Quick method for create a gate.Measure isinstance.

    Args:
        measure_frequency (Optional[float], optional): readout frequency. Defaults to None.
        measure_duration (Optional[float], optional): readout duration. Defaults to None.
        measure_amp (Optional[float], optional): [description]. readout amplitude to None.
        measure_phi (float, optional): svm add parameters. Defaults to 0.
        measure_threshold (float, optional): svm add parameters. Defaults to 0.
        measure_type (str, optional): readout type. Defaults to "default".

    Returns:
        dict: keys to be completed.
    """
    return {
        "params": {
            "frequency": measure_frequency,
            "duration": measure_duration,
            "amp": measure_amp,
            "phi": measure_phi,
            "threshold": measure_threshold,
            "bias": measure_bias,
        },
        "default_type": measure_type
    }


def create_gate_rfUnitary(default_type: str = "default", rfUnitary_shape: str = "CosPulse", rfUnitary_frequency: Optional[float] = None,
                          rfUnitary_amp: list[list[float]] = [[0, 0.5], [0, 1]], rfUnitary_duration: list[list[float]] = [[0, 0.5], [20e-9, 20e-9]],
                          rfUnitary_phase: list[list[float]] = [[-1, 1], [-1, 1]],  rfUnitary_buffer: float = 0,
                          rfUnitary_alpha: float = 1, rfUnitary_beta: float = 0, rfUnitary_delta: float = 0, **kw):
    """
    Quick method for create a gate.rfUnitary isinstance.

    Args:
        rfUnitary_type (str, optional): rfUnitary type. Defaults to "default".
        rfUnitary_shape (str, optional): rfUnitary shape. Defaults to "CosPulse".
        rfUnitary_frequency (Optional[float], optional): rfUnitary frequency Defaults to None.
        rfUnitary_amp (list[list[float]], optional): rfUnitary amplitude. Defaults to [[0, 0.5], [0, 1]].
        rfUnitary_duration (list[list[float]], optional): rfUnitary duration. Defaults to [[0, 0.5], [20e-9, 20e-9]].
        rfUnitary_phase (list[list[float]], optional): rfUnitary phase. Defaults to [[-1, 1], [-1, 1]].
        rfUnitary_alpha (float, optional): rfUnitary DRAG alpha. Defaults to 1.
        rfUnitary_beta (float, optional): rfUnitary beta alpha. Defaults to 0.
        rfUnitary_delta (float, optional): rfUnitary delta alpha. Defaults to 0.

    Returns:
        dict: keys to be completed.
    """
    return {
        "default_type": default_type,
        "params": {
            "shape": rfUnitary_shape,
            "frequency": rfUnitary_frequency,
            "amp": rfUnitary_amp,
            "duration": rfUnitary_duration,
            "phase": rfUnitary_phase,
            "alpha": rfUnitary_alpha,
            "beta": rfUnitary_beta,
            "delta": rfUnitary_delta,
            "buffer": rfUnitary_buffer,
        },
    }


def create_gate_CR(default_type: str = "default", cr_duration: float = 1e-6, cr_edge_type: str = 'cos',
                   cr_edge: float = 20e-9, cr_amp1: float = 0.8, cr_amp2: float = 0, cr_drag: float = 0, cr_skew: float = 0,
                   cr_global_phase: float = 0,  cr_relative_phase: float = 0, cr_phi1: float = 0, cr_phi2: float = 0,
                   cr_buffer: float = 10e-9, **kw) -> dict:
    """
    Quick method for create a gate._CR isinstance.

    Args:
        default_type (str, optional): Type. Defaults to "default".
        cr_duration (float, optional): Duration. Defaults to 1e-6.
        cr_edge_type (str, optional): Edge type. Defaults to 'cos'.
        cr_edge (float, optional): Edge length. Defaults to 20e-9.
        cr_amp1 (float, optional): Control amplitude. Defaults to 0.8.
        cr_amp2 (float, optional): Target amplitude. Defaults to 0.
        cr_drag (float, optional): DRAG. Defaults to 0.
        cr_skew (float, optional): Skew. Defaults to 0.
        cr_global_phase (float, optional): Global phase. Defaults to 0.
        cr_relative_phase (float, optional): Relative phase. Defaults to 0.
        cr_phi1 (float, optional): Control phase. Defaults to 0.
        cr_phi2 (float, optional): Target Phase. Defaults to 0.
        cr_buffer (float, optional): Buffer. Defaults to 10e-9.

    Returns:
        dict: keys to be completed.
    """

    return {
        "default_type": default_type,
        "params": {
            "duration": cr_duration,
            "edge_type": cr_edge_type,
            "edge": cr_edge,
            "amp1": cr_amp1,
            "amp2": cr_amp2,
            "drag": cr_drag,
            "skew": cr_skew,
            "global_phase": cr_global_phase,
            "relative_phase": cr_relative_phase,
            "phi1": cr_phi1,
            "phi2": cr_phi2,
            "buffer": cr_buffer
        }
    }


def create_jpa() -> dict:
    pass


gate_mapping_1 = {
    'rfUnitary': create_gate_rfUnitary,
    'Measure': create_gate_measure,
}

gate_mapping_2 = {
    'CR': [('D', create_gate_CR)],
}



# Author: Liu Pei  <liupei200546@163.com>  2021/10/13

from abc import abstractmethod

from scipy.optimize import OptimizeResult
from waveforms.scan_iter import OptimizerConfig

class FeedAgent(object):
    """
    Agent for feedback measurement.
    """

    config = []

    def __init__(self, variable: list[dict], *args, **kwds):
        pass

    @abstractmethod
    def tell(self, x, y) -> OptimizeResult:
        raise NotImplementedError()

    @abstractmethod
    def ask(self):
        raise NotImplementedError()

    @abstractmethod
    def check(self):
        raise NotImplementedError()

    @abstractmethod
    def get_result(self) -> OptimizeResult:
        raise NotImplementedError()



from setuptools import find_packages, setup

with open('requirements.txt', encoding='utf-8') as f:
    requirements = f.read()

setup(
    name="qos_tools",
    version='0.0.2',
    author="baqis",
    author_email="baqis@baqis.ac.cn",
    url="https://gitee.com/",
    license = "MIT",
    keywords="tools for quantum lab",
    description="control, measure and visualization",
    long_description='long_description',
    long_description_content_type='text/markdown',
    packages = find_packages(),
    include_package_data = True,
    install_requires=requirements,
    python_requires='>=3.10.0',
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Natural Language :: Chinese (Simplified)',
        'Natural Language :: English',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Python :: 3.9',
        'Topic :: Scientific/Engineering :: Physics',
    ],
    project_urls={
        'source': 'https://gitee.com',
    },
)





# Author: lizhiyuan  <lizhiyuan@baqis.ac.cn>  2021/01/01

from qos_tools.core._cm import ConsManager

from qos_tools.core._async_utils import *
from qos_tools.core._dict_utils import *
from qos_tools.core._equal_utils import *
from qos_tools.core._d2c import *

from qos_tools.core._iter import CartesianSweepIterator
from qos_tools.core._cfgcontext import cfgsave,cfgload,cfgContext

from qos_tools.core._graph_func import *

from qos_tools.core._dag_cons_manager import DagConsManager


# Author: lizhiyuan  <lizhiyuan@baqis.ac.cn>  2021/01/01
# rebuild  2021/08/19

import queue
import threading
from qos_tools.core._equal_utils import is_equal
from qos_tools.core._dict_utils import flatten_dict
from qos_tools.core._graph_func import get_subgraph, extract_graph
from graphlib import TopologicalSorter
import asyncio
import logging
log = logging.getLogger(__name__)


def default_worker(task_queue, result_queue, executor=None):
    '''
    task_queue/result_queue: queue.Queue的实例化对象
    executor: 执行器
    '''
    while True:
        if task_queue.qsize():
            task_func, args, kwds = task_queue.get()
            if executor is None:
                result = task_func(*args, **kwds)
            else:
                # TODO：根据执行器接口定义
                raise
            task_queue.task_done()
            result_queue.put(result)


def creat_task(deps, func, targets, cons_manager):
    def task_func(*args, **kwds):
        if None not in args:
            if asyncio.iscoroutinefunction(func):
                tsk = func(*args)
                value = asyncio.run(tsk)
            else:
                value = func(*args)
            return dict(zip(targets, value))
        else:  # 这里只打出警告：一些依赖变量的值为None，打印出所有的键值对
            log.warning(f'{__name__} WARNING: ' +
                        ', '.join([f'{k} is {v}' for k, v in zip(deps, args)]))
            try:
                if asyncio.iscoroutinefunction(func):
                    tsk = func(*args)
                    value = asyncio.run(tsk)
                else:
                    value = func(*args)
                return dict(zip(targets, value))
            except Exception as e:
                log.warning(f'{__name__} WARNING: '+str(e))
                return {}

    args = tuple(cons_manager.get(dep) for dep in deps)
    kwds = dict()

    task_flag = (deps, targets)
    task = (task_func, args, kwds)
    return task_flag, task


class DagConsManager(object):
    '''
    约束管理器

    按照给定的约束规则，管理一个容器内部的约束关联
    '''

    def __init__(self, container, constrains=[], delimiter='.', executor=None):
        '''
        Parameters:
            container: object
                一个类字典的树状容器，每一级都支持方括号索引的形式设置读取，或支持set/get方法设置读取

            constrains: list
                约束条目，每个元素都为一个三元元组（deps, func, targets），例如，
                [
                    (('k1.k11','k1.k12'),   (lambda v1,v2:v1+v2),   'k2'       ),
                    (('k3','k4'),           (lambda v1,v2:(v1,v2)), ('k5','k6')),
                ]
                deps，元组
                func，函数或可eval的函数表达式字符串
                targets，字符串或元组，于func输出格式一致

            delimiter: str, optional
                分割符，默认 .

            executor: 执行器
                接口参考default_worker
        '''
        self.container = container
        self.delimiter = delimiter

        constrains_dict = {}
        tar_dict = {}
        tar_list = []
        for deps, func, targets in constrains:
            if isinstance(func, str):  # 如果约束的 func 是字符串，则 eval
                func = eval(func)
            if isinstance(targets, str):
                targets = (targets,)
                func_ = lambda *args, **kw: (func(*args, **kw),)
            else:
                func_ = func
            constrains_dict[(deps, targets)] = (deps, func_, targets)
            for tar in targets:
                tar_dict[tar] = (deps, targets)
                assert tar not in tar_list, f"The target '{tar}' appears repeatedly in multiple constrain items"
                tar_list.append(tar)

        self.constrains_dict = constrains_dict
        self.tar_dict = tar_dict
        self.graph_forward, self.graph_backward = extract_graph(constrains)
        ts = TopologicalSorter(self.graph_forward)
        self.static_order = list(ts.static_order())

        # turn-on the worker thread
        self.task_queue = queue.Queue()
        self.result_queue = queue.Queue()
        self.Thread = threading.Thread(target=default_worker, args=(
            self.task_queue, self.result_queue, executor), daemon=True)
        self.Thread.start()

    def _set(self, key, value, check=True):
        '''根据关键词连接的字符串，设置对应的值

        Parameters:
            key: str
                使用符号串联关键词的字符串，比如'pulse.width.value'

            value: 
                待设入的值

            check: bool
                是否检查设入值与原值相等
        Return: 
            bool, True 表示value改变并且已设置，False 表示未变
        '''
        ks = key.rsplit(self.delimiter, maxsplit=1)
        if len(ks) > 1:
            _ks, _k = ks[0], ks[1]
            _ct = self.get(_ks)
        else:
            _k = key
            _ct = self.container
        assert _ct is not None, f"key '{key}' not exist !"

        if check and is_equal(value, _ct.get(_k)):
            return False
        else:
            try:
                _ct[_k] = value
            except TypeError:  # object is not subscriptable
                _ct.set(_k, value)
            except Exception as e:
                raise e
            return True

    def set(self, key, value, check=True):
        is_changed = self._set(key, value, check=check)
        if is_changed:
            self.exec_dag([key])
        return is_changed

    def get(self, key, default=None):
        '''根据关键词连接的字符串，获取对应的值

        Parameters:
            key: str
                使用符号串联关键词的字符串，比如'pulse.width.value'

            default: 
                类似于字典方法，当关键词不存在时返回一个默认值

        Return:
            返回键值串对应的值
        '''
        ks = key.split(self.delimiter)
        _v = self.container
        for k in ks:
            try:
                _v = _v[k]
            except KeyError:
                _v = default
            except TypeError:  # object is not subscriptable
                _v = _v.get(k)
            except Exception as e:
                raise e
        return _v

    def update(self, d, check=True):
        '''批量更新信息

        Parameters:
            d: dict
                待更新信息的字典

            check: bool,optional
                是否检查设入值与原值相等
        '''
        keys_changed = []
        d_flatten = flatten_dict(d, self.delimiter)
        for k, v in d_flatten.items():
            is_changed = self._set(k, v, check=check)
            if is_changed:
                keys_changed.append(k)
        self.exec_dag(keys_changed)

    def exec_dag(self, keys_changed):
        '''根据传入的节点列表，通过有向图的方式执行约束函数
        keys_changed: 发生了变化的节点的列表
        '''
        subgraph_nodes = set(keys_changed) & set(self.static_order)
        subgraph = get_subgraph(
            subgraph_nodes, graph_backward=self.graph_backward)
        ts = TopologicalSorter(subgraph)
        ts.prepare()
        count = 0
        while ts.is_active():
            task_dict = {}
            for node in ts.get_ready():
                if count == 0:
                    ts.done(node)
                else:
                    deps, func, targets = self.constrains_dict[self.tar_dict[node]]
                    task_flag, task = creat_task(deps, func, targets, self)
                    task_dict[task_flag] = task
            for task in task_dict.values():
                self.task_queue.put(task)

            if self.result_queue.qsize():
                result = self.result_queue.get()
                self.result_queue.task_done()
                for k, v in result.items():
                    self._set(k, v, check=False)
                    ts.done(k)
            count += 1
